""" Modules for gradients """
import torch
from torch.autograd import Function
from torch.nn import Module
import torch.nn.functional as F


def relu_grad(preact):
    return (preact > 0).float()


class SqrOp(Function):
    @staticmethod
    def symbolic(_, x):
        return _.op('grad::Sqr', x)

    @staticmethod
    def forward(ctx, x):
        return torch.square(x)


class Conv2dGrad(Module):
    def __init__(self, fw_module, weight, stride, padding, dilation, groups):
        super().__init__()
        self.weight = weight
        self.dilation = dilation
        self.groups = groups
        self.fw_module = fw_module

        assert isinstance(stride, list) and stride[0] == stride[1]
        assert isinstance(padding, list) and padding[0] == padding[1]
        assert isinstance(dilation, list) and dilation[0] == dilation[1]
        self.stride = stride[0]
        self.padding = padding[0]
        self.dilation = dilation[0]

    def forward(self, grad_last):
        output_padding0 = (
            int(self.fw_module.input_shape[2])
            - (int(self.fw_module.output_shape[2]) - 1) * self.stride
            + 2 * self.padding - 1 - (int(self.weight.size()[2] - 1) * self.dilation))
        output_padding1 = (
            int(self.fw_module.input_shape[3])
            - (int(self.fw_module.output_shape[3]) - 1) * self.stride
            + 2 * self.padding - 1 - (int(self.weight.size()[3] - 1) * self.dilation))

        return Conv2dGradOp.apply(
            grad_last, self.weight, self.stride, self.padding, self.dilation,
            self.groups, output_padding0, output_padding1)


class LinearGrad(Module):
    def __init__(self, weight):
        super().__init__()
        self.weight = weight

    def forward(self, grad_last):
        weight = self.weight.to(grad_last).t()
        return F.linear(grad_last, weight)


class ReLUGradOp(Function):
    """ Local gradient of ReLU.

    Not including multiplication with gradients from other layers.
    """
    @staticmethod
    def symbolic(_, g, g_relu, g_relu_rev, preact):
        return _.op('grad::Relu', g, g_relu, g_relu_rev, preact).setType(g.type())

    @staticmethod
    def forward(ctx, g, g_relu, g_relu_rev, preact):
        return g * relu_grad(preact)


class ReLUGrad(Module):
    def forward(self, g, preact):
        g_relu = F.relu(g)
        g_relu_rev = -F.relu(-g)
        return ReLUGradOp.apply(g, g_relu, g_relu_rev, preact)


class ReshapeGrad(Module):
    def forward(self, grad_last, inp):
        if grad_last.numel() == inp.numel():
            return grad_last.reshape(grad_last.shape[0], *inp.shape[1:])
        else:
            return grad_last.reshape(*grad_last.shape[:2], *inp.shape[1:])


class Conv2dGradOp(Function):
    @staticmethod
    def symbolic(
            g, x, w, stride, padding, dilation, groups,
            output_padding0, output_padding1):
        return g.op(
            'grad::Conv2d', x, w, stride_i=stride, padding_i=padding,
            dilation_i=dilation, groups_i=groups, output_padding0_i=output_padding0,
            output_padding1_i=output_padding1).setType(x.type())

    @staticmethod
    def forward(
            ctx, grad_last, w, stride, padding, dilation, groups, output_padding0,
            output_padding1):
        grad_shape = grad_last.shape
        grad = F.conv_transpose2d(
            grad_last.view(grad_shape[0], *grad_shape[1:]), w, None,
            stride=stride, padding=padding, dilation=dilation,
            groups=groups, output_padding=(output_padding0, output_padding1))

        grad = grad.view((grad_shape[0], *grad.shape[1:]))
        return grad


class GradNorm(Module):
    def __init__(self, norm=1):
        super().__init__()
        self.norm = norm

    def forward(self, grad):
        grad = grad.view(grad.size(0), -1)
        if self.norm == 1:
            # torch.norm is not supported in auto_LiRPA yet
            # use simpler operators for now
            return grad.abs().sum(dim=-1, keepdim=True)
        elif self.norm == 2:
            return SqrOp.apply(grad).sum(dim=-1)
        else:
            raise NotImplementedError(self.norm)


class JVP(Module):
    def __init__(self, vector):
        super().__init__()
        self.vector = vector.view(-1)

    def forward(self, grad):
        grad = grad.view(grad.size(0), -1)
        return F.linear(grad, self.vector.view(1, -1).to(grad)).squeeze(-1)



"""Handle Jacobian bounds."""

import torch
import numpy as np
from .bound_ops import JacobianOP, GradNorm  # pylint: disable=unused-import
from .bound_ops import (
    BoundInput, BoundParams, BoundAdd, BoundRelu, BoundJacobianInit, JVP,
    BoundJacobianOP)
from .utils import Flatten, logger, prod, get_spec_matrix
from collections import deque

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .bound_general import BoundedModule


def _expand_jacobian(self):
    self.jacobian_start_nodes = []
    for node in list(self.nodes()):
        if isinstance(node, BoundJacobianOP):
            self.jacobian_start_nodes.append(node.inputs[0])
            expand_jacobian_node(self, node)
    if self.jacobian_start_nodes:
        # Disable unstable options
        self.bound_opts.update({
            'sparse_intermediate_bounds': False,
            'sparse_conv_intermediate_bounds': False,
            'sparse_intermediate_bounds_with_ibp': False,
            'sparse_features_alpha': False,
            'sparse_spec_alpha': False,
        })
        for node in self.nodes():
            if isinstance(node, BoundRelu):
                node.use_sparse_spec_alpha = node.use_sparse_features_alpha = False


def expand_jacobian_node(self, jacobian_node):
    """New API for converting a graph with Jacobian

    Based on the old API `augment_gradient_graph`.
    """
    logger.info(f'Expanding Jacobian node {jacobian_node}')

    output_node = jacobian_node.inputs[0]
    input_node = jacobian_node.inputs[1]
    batch_size = output_node.output_shape[0]
    output_dim = prod(output_node.output_shape[1:])

    # Gradient values in `grad` may not be accurate. We do not consider gradient
    # accumulation from multiple succeeding nodes. We only want the shapes but
    # not the accurate values.
    grad = {}
    # Dummy values in grad_start
    grad_start = torch.ones(batch_size, output_dim, output_dim, device=self.device)
    grad[output_node.name] = grad_start
    input_node_found = False

    # First BFS pass: traverse the graph, count degrees, and build gradient
    # layers.
    # Degrees of nodes.
    degree = {}
    # Original layer for gradient computation.
    layer_grad = {}
    # Input nodes in gradient computation in back propagation.
    input_nodes = {}
    # Dummy input values for gradient computation received.
    grad_input = {}
    # Extra nodes as arguments used for gradient computation.
    # They must match the order in grad_input.
    grad_extra_nodes = {}

    degree[output_node.name] = 0
    queue = deque([output_node])
    while len(queue) > 0:
        node = queue.popleft()
        grad_extra_nodes[node.name] = []
        input_nodes[node.name] = node.inputs

        if node == input_node:
            input_node_found = True
            layer_grad[node.name] = Flatten()
            grad_input[node.name] = (grad[node.name],)
        else:
            ret = node.build_gradient_node(grad[node.name])
            node_grad, grad_input_, grad_extra_nodes_ = ret
            layer_grad[node.name] = node_grad
            grad_input[node.name] = grad_input_
            grad_extra_nodes[node.name] = grad_extra_nodes_

        # Propagate gradients to the input nodes and update degrees.
        grad_next = layer_grad[node.name](*grad_input[node.name])
        if isinstance(grad_next, torch.Tensor):
            grad_next = [grad_next]
        if not isinstance(node, BoundInput):
            for i in range(len(grad_next)):
                grad[input_nodes[node.name][i].name] = grad_next[i]
                if not input_nodes[node.name][i].name in degree:
                    degree[input_nodes[node.name][i].name] = 0
                    queue.append(input_nodes[node.name][i])
                degree[input_nodes[node.name][i].name] += 1

    if not input_node_found:
        raise RuntimeError('Input node not found')

    # Second BFS pass: build the backward computational graph
    grad_node = {}
    initial_name = f'/jacobian{output_node.name}{output_node.name}'
    grad_node[output_node.name] = BoundJacobianInit(inputs=[output_node])
    grad_node[output_node.name].name = initial_name
    self.add_nodes([grad_node[output_node.name]])
    queue = deque([output_node])
    while len(queue) > 0:
        node = queue.popleft()
        nodes_op, nodes_in, nodes_out, _ = self._convert_nodes(
            layer_grad[node.name], grad_input[node.name])
        rename_dict = {}
        assert isinstance(nodes_in[0], BoundInput)
        rename_dict[nodes_in[0].name] = grad_node[node.name].name
        for i in range(1, len(nodes_in)):
            # Assume it's a parameter here
            new_name = f'/jacobian{output_node.name}{node.name}/params{nodes_in[i].name}'
            rename_dict[nodes_in[i].name] = new_name
        for i in range(len(nodes_op)):
            # intermediate nodes
            if not nodes_op[i].name in rename_dict:
                new_name = f'/jacobian{output_node.name}{node.name}/tmp{nodes_op[i].name}'
                rename_dict[nodes_op[i].name] = new_name
        if node == input_node:
            assert len(nodes_out) == 1
            rename_dict[nodes_out[0].name] = f'/jacobian{output_node.name}'
        else:
            for i in range(len(nodes_out)):
                assert not isinstance(node.inputs[i], BoundParams)
                rename_dict[nodes_out[i].name] = f'/jacobian{output_node.name}{node.inputs[i].name}'

        self.rename_nodes(nodes_op, nodes_in, rename_dict)
        # Replace input nodes
        # grad_extra_nodes[node.name]: ReLU's input
        input_nodes_replace = (
            [self._modules[nodes_in[0].name]] + grad_extra_nodes[node.name])
        for i in range(len(input_nodes_replace)):
            for n in nodes_op:
                for j in range(len(n.inputs)):
                    if n.inputs[j].name == nodes_in[i].name:
                        n.inputs[j] = input_nodes_replace[i]
        self.add_nodes(nodes_op + nodes_in[len(input_nodes_replace):])

        if node != input_node:
            for i in range(len(nodes_out)):
                if input_nodes[node.name][i].name in grad_node:
                    node_cur = grad_node[input_nodes[node.name][0].name]
                    node_add = BoundAdd(
                        attr=None, inputs=[node_cur, nodes_out[i]],
                        output_index=0, options={})
                    node_add.name = f'{nodes_out[i].name}/add'
                    grad_node[input_nodes[node.name][0].name] = node_add
                else:
                    grad_node[input_nodes[node.name][0].name] = nodes_out[i]
                degree[input_nodes[node.name][i].name] -= 1
                if degree[input_nodes[node.name][i].name] == 0:
                    queue.append(input_nodes[node.name][i])
        else:
            self.replace_node(jacobian_node, grad_node[node.name])


def augment_gradient_graph(self: 'BoundedModule', dummy_input, norm=None,
                           vector=None):
    """Augment the computational graph with gradient computation."""
    device = dummy_input.device
    final_node = self.final_node()
    if final_node.forward_value is None:
        self(dummy_input)
    output = final_node.forward_value
    if output.ndim != 2:
        raise NotImplementedError(
            'The model should have a 2-D output shape of '
            '(batch_size, output_dim)')
    output_dim = output.size(1)

    # Gradient values in `grad` may not be accurate. We do not consider gradient
    # accumulation from multiple succeeding nodes. We only want the shapes but
    # not the accurate values.
    grad = {}
    # Dummy values in grad_start
    grad_start = torch.ones(output.size(0), output_dim, device=device)
    grad[final_node.name] = grad_start
    input_node_found = False

    # First BFS pass: traverse the graph, count degrees, and build gradient
    # layers.
    # Degrees of nodes.
    degree = {}
    # Original layer for gradient computation.
    layer_grad = {}
    # Input nodes in gradient computation in back propagation.
    input_nodes = {}
    # Dummy input values for gradient computation received.
    grad_input = {}
    # Extra nodes as arguments used for gradient computation.
    # They must match the order in grad_input.
    grad_extra_nodes = {}

    degree[final_node.name] = 0
    queue = deque([final_node])
    while len(queue) > 0:
        node = queue.popleft()
        grad_extra_nodes[node.name] = []
        input_nodes[node.name] = node.inputs

        if isinstance(node, BoundInput):
            if input_node_found:
                raise NotImplementedError(
                    'There must be exactly one BoundInput node, '
                    'but found more than 1.')
            if vector is not None:
                # Compute JVP bounds
                layer_grad[node.name] = JVP(vector=vector)
            else:
                if norm is None:
                    layer_grad[node.name] = Flatten()
                else:
                    dual_norm = 1. / (1. - 1. / norm) if norm != 1 else np.inf
                    layer_grad[node.name] = GradNorm(norm=dual_norm)
            grad_input[node.name] = (grad[node.name],)
            input_node_found = True
        else:
            ret = node.build_gradient_node(grad[node.name])
            node_grad, grad_input_, grad_extra_nodes_ = ret
            layer_grad[node.name] = node_grad
            grad_input[node.name] = grad_input_
            grad_extra_nodes[node.name] = grad_extra_nodes_

        # Propagate gradients to the input nodes and update degrees.
        grad_next = layer_grad[node.name](*grad_input[node.name])
        if isinstance(grad_next, torch.Tensor):
            grad_next = [grad_next]
        if not isinstance(node, BoundInput):
            for i in range(len(grad_next)):
                grad[input_nodes[node.name][i].name] = grad_next[i]
                if not input_nodes[node.name][i].name in degree:
                    degree[input_nodes[node.name][i].name] = 0
                    queue.append(input_nodes[node.name][i])
                degree[input_nodes[node.name][i].name] += 1

    if not input_node_found:
        raise NotImplementedError(
            'There must be exactly one BoundInput node, but found none.')

    # Second BFS pass: build the backward computational graph
    grad_node = {}
    grad_node[final_node.name] = BoundInput(
        f'/grad{final_node.name}', grad_start)
    grad_node[final_node.name].name = f'/grad{final_node.name}'
    self.add_input_node(grad_node[final_node.name], index='auto')
    queue = deque([final_node])
    while len(queue) > 0:
        node = queue.popleft()
        nodes_op, nodes_in, nodes_out, _ = self._convert_nodes(
            layer_grad[node.name], grad_input[node.name])
        rename_dict = {}
        assert isinstance(nodes_in[0], BoundInput)
        rename_dict[nodes_in[0].name] = grad_node[node.name].name
        for i in range(1, len(nodes_in)):
            # Assume it's a parameter here
            new_name = f'/grad{node.name}/params{nodes_in[i].name}'
            rename_dict[nodes_in[i].name] = new_name
        for i in range(len(nodes_op)):
            # intermediate nodes
            if not nodes_op[i].name in rename_dict:
                new_name = f'/grad{node.name}/tmp{nodes_op[i].name}'
                rename_dict[nodes_op[i].name] = new_name
        if isinstance(node, BoundInput):
            assert len(nodes_out) == 1
            rename_dict[nodes_out[0].name] = '/grad_norm'
        else:
            for i in range(len(nodes_out)):
                assert not isinstance(node.inputs[i], BoundParams)
                rename_dict[nodes_out[i].name] = f'/grad{node.inputs[i].name}'

        self.rename_nodes(nodes_op, nodes_in, rename_dict)
        # Replace input nodes
        # grad_extra_nodes[node.name]: ReLU's input
        input_nodes_replace = (
            [self._modules[nodes_in[0].name]] + grad_extra_nodes[node.name])
        for i in range(len(input_nodes_replace)):
            for n in nodes_op:
                for j in range(len(n.inputs)):
                    if n.inputs[j].name == nodes_in[i].name:
                        n.inputs[j] = input_nodes_replace[i]
        self.add_nodes(nodes_op + nodes_in[len(input_nodes_replace):])

        if not isinstance(node, BoundInput):
            for i in range(len(nodes_out)):
                if input_nodes[node.name][i].name in grad_node:
                    node_cur = grad_node[input_nodes[node.name][0].name]
                    node_add = BoundAdd(
                        attr=None, inputs=[node_cur, nodes_out[i]],
                        output_index=0, options={})
                    node_add.name = f'{nodes_out[i].name}/add'
                    grad_node[input_nodes[node.name][0].name] = node_add
                else:
                    grad_node[input_nodes[node.name][0].name] = nodes_out[i]
                degree[input_nodes[node.name][i].name] -= 1
                if degree[input_nodes[node.name][i].name] == 0:
                    queue.append(input_nodes[node.name][i])

    self(dummy_input, grad_start, final_node_name='/grad_norm')

    self.bound_opts['jacobian'] = {
        'norm': norm,
        'vector': vector is not None,
    }
    # Disable unstable options
    self.bound_opts.update({
        'sparse_intermediate_bounds': False,
        'sparse_conv_intermediate_bounds': False,
        'sparse_intermediate_bounds_with_ibp': False,
        'sparse_features_alpha': False,
        'sparse_spec_alpha': False,
    })
    for node in self.nodes():
        if isinstance(node, BoundRelu):
            node.use_sparse_spec_alpha = node.use_sparse_features_alpha = False

    self.forward_final_name = self.final_name
    self.final_name = '/grad_norm'
    self.output_name = ['/grad_norm']

    return self


def compute_jacobian_bounds(self: 'BoundedModule', x, optimize=True,
                            reduce=True, c_opt=None, labels=None):
    """Compute jacobian bounds on the pre-augmented graph.

    Args:
        x: Input to the mode.
        optimize: Optimize relaxation bounds.
        reduce: Reduce the Jacobian bounds by taking the max
        (for L-inf local Lipschitz constants).
        c_opt (optional): Specification for optimizing the bounds on the forward
        computational graph.
        labels (optional): Ground-truth labels for constructing a default c_opt
        if it c_opt not explicitly provided.

    Returns:
        ret: If reduce=True, return the maximum Jacobian bounds over all the
        output dimensions, otherwise return the Jacobian bounds for all the
        output dimensions in a tensor.
    """
    assert 'jacobian' in self.bound_opts, (
        'Call augment_gradient_graph to augment the computational graph '
        'with the backward graph first')
    norm = self.bound_opts.get('jacobian', {}).get('norm', None)

    num_classes = self[self.forward_final_name].output_shape[-1]

    if optimize and c_opt is None:
        if labels is None:
            c_opt = None
        else:
            # Specification for optimizing the forward graph.
            c_opt = get_spec_matrix(x, labels, num_classes)

    ret, lower, upper = [], [], []
    grad_start = torch.zeros(x.size(0), num_classes).to(x)
    for j in range(num_classes):
        grad_start.zero_()
        grad_start[:, j] = 1
        x_extra = (grad_start,)
        intermediate_bounds = {}
        if optimize:
            self.compute_bounds(
                method='CROWN-Optimized',
                C=c_opt, x=(x,) + x_extra, bound_upper=False,
                final_node_name=self.forward_final_name)
            for node in self._modules.values():
                if hasattr(node, 'lower') and node.lower is not None:
                    intermediate_bounds[node.name] = (node.lower, node.upper)
        lb, ub = self.compute_bounds(
                method='CROWN', x=(x,) + x_extra, bound_lower=norm is None,
                interm_bounds=intermediate_bounds)
        if norm is not None:
            ret.append(ub.view(-1))
        else:
            lower.append(lb.view(1, -1))
            upper.append(ub.view(1, -1))
    if norm is not None:
        ret = torch.concat(ret)
        if reduce:
            return ret.max()
        else:
            return ret
    else:
        lower = torch.concat(lower, dim=0)
        upper = torch.concat(upper, dim=0)
        return lower, upper


def compute_jacobian_bounds_new(self: 'BoundedModule', x, optimize=True,
                            optimize_output_node=None,
                            bound_lower=True, bound_upper=True):
    """Compute jacobian bounds on the pre-augmented graph (new API)."""

    if isinstance(x, torch.Tensor):
        x = (x,)

    if optimize:
        if optimize_output_node is None:
            if len(self.jacobian_start_nodes) == 1:
                optimize_output_node = self.jacobian_start_nodes[0]
            else:
                raise NotImplementedError(
                    'Multiple Jacobian nodes found.'
                    'An output node for optimizable bounds (optimize_output_node) '
                    'must be specified explicitly')
        self.compute_bounds(
            method='CROWN-Optimized',
            C=None, x=x, bound_upper=False,
            final_node_name=optimize_output_node.name)
        intermediate_bounds = {}
        for node in self._modules.values():
            if hasattr(node, 'lower') and node.lower is not None:
                intermediate_bounds[node.name] = (node.lower, node.upper)
    else:
        intermediate_bounds = None
    lb, ub = self.compute_bounds(
        method='CROWN', x=x,
        bound_lower=bound_lower, bound_upper=bound_upper,
        interm_bounds=intermediate_bounds)
    return lb, ub



import torch

from ..operations.base import Operator
from ..utils import get_selection


class Squeeze(Operator):
    def __init__(self, opset_version, dim=None):
        self.opset_version = opset_version
        self.dim = dim
        super().__init__()

    def forward(self, input: torch.Tensor, axes: torch.Tensor = None):
        if self.opset_version < 13:
            dims = self.dim
        else:
            dims = axes

        if dims is None:
            return torch.squeeze(input)
        elif isinstance(dims, int):
            return torch.squeeze(input, dim=dims)
        else:
            for dim in sorted(dims, reverse=True):
                input = torch.squeeze(input, dim=dim)
            return input




import warnings
warnings.filterwarnings(action='ignore')
import unittest
import logging

from .verifier.objective import Objective, DnfObjectives
from .util.spec.read_vnnlib import read_vnnlib
from .util.network.read_onnx import parse_onnx
from .verifier.verifier import Verifier
from .util.misc.result import ReturnStatus
from .util.misc.logger import logger
from .setting import Settings



def extract_instance(net_path, vnnlib_path):
    vnnlibs = read_vnnlib(vnnlib_path)
    model, input_shape, output_shape, is_nhwc = parse_onnx(net_path)

    # objective
    objectives = []
    for spec in vnnlibs:
        bounds = spec[0]
        for prop_i in spec[1]:
            objectives.append(Objective((bounds, prop_i)))
    objectives = DnfObjectives(objectives, input_shape=input_shape, is_nhwc=is_nhwc)

    return model, input_shape, objectives

def reset_settings():
    Settings.__init__()
    Settings.setup(args=None)


class TestVerifier(unittest.TestCase):

    def test_mnist_sat(self):
        reset_settings()
        net_path = 'example/onnx/mnist-net_256x2.onnx'
        vnnlib_path = 'example/vnnlib/prop_1_0.05.vnnlib'
        device = 'cuda'

        print('\n\nRunning test with', net_path, vnnlib_path)

        model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
        model.to(device)

        verifier = Verifier(
            net=model,
            input_shape=input_shape,
            batch=1000,
            device=device,
        )

        status = verifier.verify(objectives)

        self.assertEqual(status, ReturnStatus.SAT)


    def test_mnist_gdvb_unsat(self):
        reset_settings()
        net_path = 'example/onnx/mnistfc-medium-net-151.onnx'
        vnnlib_path = 'example/vnnlib/prop_2_0.03.vnnlib'
        device = 'cuda'

        print('\n\nRunning test with', net_path, vnnlib_path)

        model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
        model.to(device)

        verifier = Verifier(
            net=model,
            input_shape=input_shape,
            batch=1000,
            device=device,
        )

        status = verifier.verify(objectives)

        self.assertEqual(status, ReturnStatus.UNSAT)



    def test_mnist_unsat(self):
        reset_settings()
        net_path = 'example/onnx/mnist-net_256x2.onnx'
        vnnlib_path = 'example/vnnlib/prop_1_0.03.vnnlib'
        device = 'cuda'

        print('\n\nRunning test with', net_path, vnnlib_path)

        model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
        model.to(device)

        verifier = Verifier(
            net=model,
            input_shape=input_shape,
            batch=1000,
            device=device,
        )

        status = verifier.verify(objectives)

        self.assertEqual(status, ReturnStatus.UNSAT)



    def test_mnist_unsat_w_restart(self):
        reset_settings()
        Settings.use_mip_tightening = False
        Settings.use_restart = True
        Settings.max_hidden_visited_branches = 100

        net_path = 'example/onnx/mnist-net_256x2.onnx'
        vnnlib_path = 'example/vnnlib/prop_1_0.03.vnnlib'
        device = 'cuda'

        print('\n\nRunning test with', net_path, vnnlib_path)

        model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
        model.to(device)

        verifier = Verifier(
            net=model,
            input_shape=input_shape,
            batch=1000,
            device=device,
        )

        status = verifier.verify(objectives)

        self.assertEqual(status, ReturnStatus.UNSAT)


    def test_mnist_unsat_w_stablize(self):
        reset_settings()
        Settings.use_mip_tightening = True
        Settings.use_restart = False
        Settings.mip_tightening_timeout_per_neuron = 2.0

        net_path = 'example/onnx/mnist-net_256x2.onnx'
        vnnlib_path = 'example/vnnlib/prop_1_0.03.vnnlib'
        device = 'cuda'

        print('\n\nRunning test with', net_path, vnnlib_path)

        model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
        model.to(device)

        verifier = Verifier(
            net=model,
            input_shape=input_shape,
            batch=1000,
            device=device,
        )

        status = verifier.verify(objectives)

        self.assertEqual(status, ReturnStatus.UNSAT)


    def test_mnist_unsat_w_stablize_and_restart(self):
        reset_settings()
        Settings.use_mip_tightening = True
        Settings.use_restart = True
        Settings.mip_tightening_timeout_per_neuron = 2.0
        Settings.max_hidden_visited_branches = 100

        net_path = 'example/onnx/mnist-net_256x2.onnx'
        vnnlib_path = 'example/vnnlib/prop_1_0.03.vnnlib'
        device = 'cuda'

        print('\n\nRunning test with', net_path, vnnlib_path)

        model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
        model.to(device)

        verifier = Verifier(
            net=model,
            input_shape=input_shape,
            batch=1000,
            device=device,
        )

        status = verifier.verify(objectives)

        self.assertEqual(status, ReturnStatus.UNSAT)

    def test_acas_unsat(self):
        reset_settings()
        net_path = 'example/onnx/ACASXU_run2a_1_1_batch_2000.onnx'
        vnnlib_path = 'example/vnnlib/prop_6.vnnlib'
        device = 'cuda'

        print('\n\nRunning test with', net_path, vnnlib_path)

        model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
        model.to(device)

        verifier = Verifier(
            net=model,
            input_shape=input_shape,
            batch=1000,
            device=device,
        )

        status = verifier.verify(objectives)

        self.assertEqual(status, ReturnStatus.UNSAT)


    def test_acas_sat(self):
        reset_settings()
        net_path = 'example/onnx/ACASXU_run2a_1_9_batch_2000.onnx'
        vnnlib_path = 'example/vnnlib/prop_7.vnnlib'
        device = 'cuda'

        print('\n\nRunning test with', net_path, vnnlib_path)

        model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
        model.to(device)

        verifier = Verifier(
            net=model,
            input_shape=input_shape,
            batch=1000,
            device=device,
        )

        status = verifier.verify(objectives)

        self.assertEqual(status, ReturnStatus.SAT)


    def test_nn4sys_unsat_1(self):
        reset_settings()
        net_path = 'example/onnx/mscn_128d.onnx'
        vnnlib_path = 'example/vnnlib/cardinality_0_100_128.vnnlib'
        device = 'cuda'

        print('\n\nRunning test with', net_path, vnnlib_path)

        model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
        model.to(device)

        verifier = Verifier(
            net=model,
            input_shape=input_shape,
            batch=1000,
            device=device,
        )

        status = verifier.verify(objectives)

        self.assertEqual(status, ReturnStatus.UNSAT)



    def test_nn4sys_unsat_2(self):
        reset_settings()
        net_path = 'example/onnx/pensieve_big_parallel.onnx'
        vnnlib_path = 'example/vnnlib/pensieve_parallel_55.vnnlib'
        device = 'cuda'

        print('\n\nRunning test with', net_path, vnnlib_path)

        model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
        model.to(device)

        verifier = Verifier(
            net=model,
            input_shape=input_shape,
            batch=1000,
            device=device,
        )

        status = verifier.verify(objectives)

        self.assertEqual(status, ReturnStatus.UNSAT)


    def test_cifar_unsat_1(self):
        reset_settings()
        net_path = 'example/onnx/cifar10_2_255_simplified.onnx'
        vnnlib_path = 'example/vnnlib/cifar10_spec_idx_4_eps_0.00784_n1.vnnlib'
        device = 'cuda'

        print('\n\nRunning test with', net_path, vnnlib_path)

        model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
        model.to(device)

        verifier = Verifier(
            net=model,
            input_shape=input_shape,
            batch=200,
            device=device,
        )

        status = verifier.verify(objectives)

        self.assertEqual(status, ReturnStatus.UNSAT)



    def test_cifar_unsat_2(self):
        reset_settings()
        net_path = 'example/onnx/convBigRELU__PGD.onnx'
        vnnlib_path = 'example/vnnlib/cifar10_spec_idx_95_eps_0.00784.vnnlib'
        device = 'cuda'

        print('\n\nRunning test with', net_path, vnnlib_path)

        model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
        model.to(device)

        verifier = Verifier(
            net=model,
            input_shape=input_shape,
            batch=200,
            device=device,
        )

        status = verifier.verify(objectives)

        self.assertEqual(status, ReturnStatus.UNSAT)


    def test_cgan_unsat(self):
        reset_settings()
        net_path = 'example/onnx/cGAN_imgSz32_nCh_1.onnx'
        vnnlib_path = 'example/vnnlib/cGAN_imgSz32_nCh_1_prop_2_input_eps_0.020_output_eps_0.025.vnnlib'
        device = 'cuda'

        print('\n\nRunning test with', net_path, vnnlib_path)

        model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
        model.to(device)

        verifier = Verifier(
            net=model,
            input_shape=input_shape,
            batch=10,
            device=device,
        )

        status = verifier.verify(objectives)

        self.assertEqual(status, ReturnStatus.UNSAT)



    def test_dist_shift_unsat(self):
        reset_settings()
        net_path = 'example/onnx/mnist_concat.onnx'
        vnnlib_path = 'example/vnnlib/index188_delta0.13.vnnlib'
        device = 'cuda'

        print('\n\nRunning test with', net_path, vnnlib_path)

        model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
        model.to(device)

        verifier = Verifier(
            net=model,
            input_shape=input_shape,
            batch=1000,
            device=device,
        )

        status = verifier.verify(objectives)

        self.assertEqual(status, ReturnStatus.UNSAT)



    def test_tllverifybench_unsat(self):
        reset_settings()
        net_path = 'example/onnx/tllBench_n=2_N=M=16_m=1_instance_1_1.onnx'
        vnnlib_path = 'example/vnnlib/property_N=16_1.vnnlib'
        device = 'cuda'

        print('\n\nRunning test with', net_path, vnnlib_path)

        model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
        model.to(device)

        verifier = Verifier(
            net=model,
            input_shape=input_shape,
            batch=1000,
            device=device,
        )

        status = verifier.verify(objectives)

        self.assertEqual(status, ReturnStatus.UNSAT)



    def test_vit_unsat(self):
        reset_settings()
        net_path = 'example/onnx/pgd_2_3_16.onnx'
        vnnlib_path = 'example/vnnlib/pgd_2_3_16_4021.vnnlib'
        device = 'cuda'

        print('\n\nRunning test with', net_path, vnnlib_path)

        model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
        model.to(device)

        verifier = Verifier(
            net=model,
            input_shape=input_shape,
            batch=1000,
            device=device,
        )

        status = verifier.verify(objectives)

        self.assertEqual(status, ReturnStatus.UNSAT)



if __name__ == '__main__':
    logger.setLevel(logging.INFO)
    unittest.main()



from .bound_ops import (
    Bound, BoundLinear, BoundPrimConstant, BoundGELU, BoundReluGrad,
    BoundConv2dGrad, BoundSqr, BoundJacobianOP)

bound_op_map = {
    'onnx::Gemm': BoundLinear,
    'prim::Constant': BoundPrimConstant,
    'grad::Relu': BoundReluGrad,
    'grad::Conv2d': BoundConv2dGrad,
    'grad::Sqr': BoundSqr,
    'grad::jacobian': BoundJacobianOP,
    'custom::Gelu': BoundGELU,
}

def register_custom_op(op_name: str, bound_obj: Bound) -> None:
    bound_op_map[op_name] = bound_obj

def unregister_custom_op(op_name: str) -> None:
    bound_op_map.pop(op_name)



import torch
from .base import Bound


class JacobianOP(torch.autograd.Function):
    @staticmethod
    def symbolic(g, output, input):
        return g.op('grad::jacobian', output, input).setType(output.type())

    @staticmethod
    def forward(ctx, output, input):
        output_ = output.flatten(1)
        return torch.zeros(
            output.shape[0], output_.shape[-1], *input.shape[1:],
            device=output.device)


class BoundJacobianOP(Bound):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)

    def forward(self, output, input):
        return JacobianOP.apply(output, input)


class BoundJacobianInit(Bound):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)
        self.never_perturbed = True

    def forward(self, x):
        x = x.flatten(1)
        eye = torch.eye(x.shape[-1], device=x.device)
        return eye.unsqueeze(0).repeat(x.shape[0], 1, 1)




from pathlib import Path
import torch.nn as nn
import random
import torch
import time
import os

from verifier.verifier import Verifier 
from util.spec.read_vnnlib import read_vnnlib
from util.network.read_onnx import parse_onnx, custom_quirks
from verifier.objective import Objective, DnfObjectives
from util.misc.logger import logger
from setting import Settings

from auto_LiRPA import BoundedModule
from auto_LiRPA.perturbations import PerturbationLpNorm
from auto_LiRPA import BoundedTensor
from abstractor.params import *
from auto_LiRPA.utils import stop_criterion_batch_any, multi_spec_keep_func_all, stop_criterion_all

custom_quirks['Reshape']['fix_batch_size'] = False

def new_input(x_L: torch.Tensor, x_U: torch.Tensor) -> BoundedTensor:
    return BoundedTensor(x_L, PerturbationLpNorm(x_L=x_L, x_U=x_U)).to(x_L.device)

def extract_instance(net_path, vnnlib_path):
    vnnlibs = read_vnnlib(vnnlib_path)
    model, input_shape, output_shape, is_nhwc = parse_onnx(net_path)
    
    # objective
    objectives = []
    for spec in vnnlibs:
        bounds = spec[0]
        for prop_i in spec[1]:
            objectives.append(Objective((bounds, prop_i)))
    objectives = DnfObjectives(objectives, input_shape=input_shape, is_nhwc=is_nhwc)

    return model, input_shape, objectives


if __name__ == "__main__":
    net_path = 'example/onnx/pgd_2_3_16.onnx'
    vnnlib_path = 'example/vnnlib/pgd_2_3_16_4021.vnnlib'
    model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
    print(model)
    print(custom_quirks)
    device = 'cpu'
    device = 'cuda'
    
    bound_opts = {
        'verbosity': 1, 
        'conv_mode': 'matrix', 
        'deterministic': False, 
        'sparse_features_alpha': True, # no
        'sparse_spec_alpha': True, # no
        'sparse_intermediate_bounds': False, # yes (backward)
        'crown_batch_size': 1000000000, 
        'max_crown_size': 1000000000, 
        'forward_refinement': False, 
        'dynamic_forward': False, 
        'forward_max_dim': 10000, 
        'use_full_conv_alpha': True, 
        'disable_optimization': [], 
        'fixed_reducemax_index': True, 
        'matmul': {'share_alphas': False}, # yes
        'tanh': {'loose_threshold': None}, 
        'buffers': {'no_batchdim': False}, # yes (optimized backward)
        'optimize_graph': {'optimizer': None}
    }

    net = BoundedModule(
        model=model, 
        global_input=torch.zeros(input_shape, device=device),
        bound_opts=bound_opts,
        device=device,
        verbose=0,
    )
    
    x_L = objectives.lower_bounds[0].view(input_shape).to(device)
    x_U = objectives.upper_bounds[0].view(input_shape).to(device)
    x = new_input(x_L=x_L, x_U=x_U)
    cs = objectives.cs.transpose(0, 1).to(device)
    rhs = objectives.rhs.transpose(0, 1).to(device)
    
    print(net(x))
    net.get_split_nodes(input_split=False)
    print(net.split_nodes)
    print(net.split_activations)
    print(cs.shape)
    print(rhs.shape)
    
    net.set_bound_opts({
        'optimize_bound_args': {'stop_criterion_func': stop_criterion_all(rhs)}, 'verbosity': 1,
    })
    net.set_bound_opts({
        'optimize_bound_args': {
            'enable_alpha_crown': True, 
            'enable_beta_crown': False, 
            'init_alpha': False, 
            'fix_interm_bounds': True, 
            'use_shared_alpha': False, 
            'early_stop_patience': 10, 
        }
    })

    lb, _, aux_reference_bounds = net.init_alpha(
        (x,), share_alphas=False, c=cs, bound_upper=False)
    print('initial CROWN bounds:', lb,)
    
    lb, _ = net.compute_bounds(
        x=(x,), C=cs, method='CROWN-Optimized',
        bound_upper=False, 
        aux_reference_bounds=aux_reference_bounds,
    )    
    
    print('initial CROWN-Optimized bounds:', lb,)
    


import torch
from torch import nn


class Expand(nn.Module):
    def forward(self, input: torch.Tensor, shape: torch.Tensor):
        if isinstance(shape, torch.Tensor):
            shape = shape.to(torch.int64)
        try:
            out = input.expand(torch.Size(shape))
        except RuntimeError:
            out = input * torch.ones(
                torch.Size(shape), dtype=input.dtype, device=input.device
            )
        return out



import torch
from torch import nn


class ReduceSum(nn.Module):
    def __init__(
        self, opset_version, dim=None, keepdim=True, noop_with_empty_axes=False
    ):
        self.opset_version = opset_version
        self.dim = dim
        self.keepdim = keepdim
        self.noop_with_empty_axes = noop_with_empty_axes
        super().__init__()

    def forward(self, data: torch.Tensor, axes: torch.Tensor = None):
        if self.opset_version < 13:
            dims = self.dim
        else:
            dims = axes
        if dims is None:
            if self.noop_with_empty_axes:
                return data
            else:
                dims = tuple(range(data.ndim))
        if isinstance(dims, int):
            return torch.sum(data, dim=dims, keepdim=self.keepdim)
        else:
            return torch.sum(data, dim=tuple(list(dims)), keepdim=self.keepdim)



from beartype import beartype
import torch
import time

from ...onnx2pytorch.convert.model import ConvertModel

from .util import get_loss, check_adv_multi, serialize_specs
from ...util.misc.adam_clipping import AdamClipping


# @beartype
def attack(model: ConvertModel, x: torch.Tensor, data_min: torch.Tensor, data_max: torch.Tensor,
           cs: torch.Tensor, rhs: torch.Tensor, timeout: float,
           attack_iters: int = 100, num_restarts: int = 30) -> tuple[bool, torch.Tensor | None]:
    # set all parameters without gradient, this can speedup things significantly.
    grad_status = {}
    for p in model.parameters():
        grad_status[p] = p.requires_grad
        p.requires_grad_(False)

    output = model(x).detach()

    # specifications
    serialized_conditions = serialize_specs(x, cs, rhs)

    # sanity check
    output = output[:, None, None].repeat(1, 1, len(serialized_conditions[-1][0]), 1)
    if check_adv_multi(x, output, serialized_conditions, data_max.unsqueeze(1), data_min.unsqueeze(1)):
        x = x[:, None, None]
        return True, x.repeat(1, 1, len(serialized_conditions[-1][0]), *[1]*(x.ndim - 3)).detach()

    data_min = data_min.to(x.device)
    data_max = data_max.to(x.device)
    attack_images = general_attack(
        model=model,
        X=x,
        data_min=data_min,
        data_max=data_max,
        serialized_conditions=serialized_conditions,
        attack_iters=attack_iters,
        num_restarts=num_restarts,
        use_gama=False,
        timeout=timeout,
    )

    if attack_images is None:
        attack_images = general_attack(
            model=model,
            X=x,
            data_min=data_min,
            data_max=data_max,
            serialized_conditions=serialized_conditions,
            attack_iters=attack_iters,
            num_restarts=num_restarts,
            use_gama=True,
            timeout=timeout,
        )

    if attack_images is not None:
        return True, attack_images.detach()

    # set back to original requires_grad status.
    for p in model.parameters():
        p.requires_grad_(grad_status[p])

    return False, None



# @beartype
def general_attack(model: ConvertModel, X: torch.Tensor, data_min: torch.Tensor, data_max: torch.Tensor,
                   serialized_conditions: tuple[torch.Tensor, torch.Tensor, list[list[int]]], timeout: float,
                   use_gama: bool = False, num_restarts: int = 10, attack_iters: int = 100,
                   only_replicate_restarts: bool = False) -> torch.Tensor | None:
    # hyper params
    lr_decay = 0.99
    gama_lambda = 10.0

    # shapes
    input_shape = (X.shape[0], *X.shape[2:]) if only_replicate_restarts else X.size()
    num_classes = serialized_conditions[0].shape[-1]
    num_specs = len(serialized_conditions[-1][0])
    extra_dim = (num_restarts,) if only_replicate_restarts else (num_restarts, num_specs)

    # [1, 1, num_spec, *input_shape]
    data_min = data_min.unsqueeze(1)
    data_max = data_max.unsqueeze(1)

    X_ndim_orig = X.ndim
    # shape of x: [num_example, *shape_of_x]
    X = X.view(X.shape[0], *[1] * len(extra_dim), *X.shape[1:])
    X = X.expand(-1, *extra_dim, *(-1,) * (X_ndim_orig - 1))
    extra_dim = (X.shape[1], X.shape[2])

    delta_lower_limit = data_min - X
    delta_upper_limit = data_max - X

    lr = torch.max(data_max - data_min).item() / 8
    delta = (torch.empty_like(X).uniform_() * (delta_upper_limit - delta_lower_limit) + delta_lower_limit).requires_grad_()

    # optimizer
    opt = AdamClipping(params=[delta], lr=lr)
    scheduler = torch.optim.lr_scheduler.ExponentialLR(opt, lr_decay)

    start = time.time()
    for _ in range(attack_iters):
        inputs = torch.max(torch.min((X + delta), data_max), data_min)
        output = model(inputs.view(-1, *input_shape[1:])).view(input_shape[0], *extra_dim, num_classes)

        # early stop
        if check_adv_multi(inputs, output, serialized_conditions, data_max, data_min):
            return inputs

        if time.time() - start > timeout:
            break

        origin_out = torch.softmax(model(X.reshape(-1, *input_shape[1:])), 1).view(output.shape) if use_gama else None

        loss = get_loss(origin_out, output, serialized_conditions, gama_lambda if use_gama else 0.0)
        loss.sum().backward()

        # optimize
        opt.step(clipping=True, lower_limit=delta_lower_limit, upper_limit=delta_upper_limit, sign=1)
        opt.zero_grad(set_to_none=True)
        scheduler.step()
        gama_lambda *= 0.9


    return None



""" Bound classes for gradient operators """
import torch
import torch.nn.functional as F
from ..patches import Patches, inplace_unfold
from .base import Bound, Interval
from .activation_base import BoundActivation
from .gradient_modules import relu_grad


# FIXME reuse the function from auto_LiRPA.patches
def _maybe_unfold(d_tensor, last_A):
    if d_tensor is None:
        return None

    #[batch, out_dim, in_c, in_H, in_W]
    d_shape = d_tensor.size()

    # Reshape to 4-D tensor to unfold.
    #[batch, out_dim*in_c, in_H, in_W]
    d_tensor = d_tensor.view(d_shape[0], -1, *d_shape[-2:])
    # unfold the slope matrix as patches.
    # Patch shape is [batch, out_h, out_w, out_dim*in_c, H, W).
    d_unfolded = inplace_unfold(
        d_tensor, kernel_size=last_A.patches.shape[-2:], stride=last_A.stride,
        padding=last_A.padding)
    # Reshape to [batch, out_H, out_W, out_dim, in_C, H, W]
    d_unfolded_r = d_unfolded.view(
        *d_unfolded.shape[:3], d_shape[1], *d_unfolded.shape[-2:])
    if last_A.unstable_idx is not None:
        if len(last_A.unstable_idx) == 4:
            # [batch, out_H, out_W, out_dim, in_C, H, W]
            # to [out_H, out_W, batch, out_dim, in_C, H, W]
            d_unfolded_r = d_unfolded_r.permute(1, 2, 0, 3, 4, 5, 6)
            d_unfolded_r = d_unfolded_r[
                last_A.unstable_idx[2], last_A.unstable_idx[3]]
        else:
            raise NotImplementedError
    # For sparse patches, the shape after unfold is
    # (unstable_size, batch_size, in_c, H, W).
    # For regular patches, the shape after unfold is
    # (spec, batch, out_h, out_w, in_c, H, W).
    return d_unfolded_r


class BoundReluGrad(BoundActivation):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)
        self.requires_input_bounds = [3]
        self.recurjac = options.get('recurjac', False)

    @staticmethod
    def relu_grad(preact):
        return (preact > 0).float()

    def forward(self, g, g_relu, g_relu_rev, preact):
        if g.ndim == preact.ndim + 1:
            preact = preact.unsqueeze(1)
        return g * relu_grad(preact)

    def interval_propagate(self, *v):
        g_lower, g_upper = v[0]
        preact_lower, preact_upper = v[3]
        relu_grad_lower = relu_grad(preact_lower)
        relu_grad_upper = relu_grad(preact_upper)
        if g_lower.ndim == relu_grad_lower.ndim + 1:
            relu_grad_lower = relu_grad_lower.unsqueeze(1)
            relu_grad_upper = relu_grad_upper.unsqueeze(1)
        lower = torch.min(g_lower * relu_grad_lower, g_lower * relu_grad_upper)
        upper = torch.max(g_upper * relu_grad_lower, g_upper * relu_grad_upper)
        return lower, upper

    def bound_backward(self, last_lA, last_uA, g, g_relu, g_relu_rev, preact,
                       **kwargs):
        mask_active = (preact.lower > 0).float()
        mask_inactive = (preact.upper < 0).float()
        mask_unstable = 1 - mask_active - mask_inactive

        if self.recurjac and self.inputs[0].perturbed:
            upper_grad = preact.upper >= 0
            lower_interval = self.inputs[0].lower * upper_grad
            upper_interval = self.inputs[0].upper * upper_grad
        else:
            lower_interval = upper_interval = None

        def _bound_oneside(last_A, pos_interval=None, neg_interval=None):
            if last_A is None:
                return None, None, None, 0

            if isinstance(last_A, torch.Tensor):
                if self.recurjac and self.inputs[0].perturbed:
                    mask_unstable_grad = (
                        (self.inputs[0].lower < 0) * (self.inputs[0].upper > 0))
                    last_A_unstable = last_A * mask_unstable_grad
                    bias = (
                        last_A_unstable.clamp(min=0) * pos_interval
                        + last_A_unstable.clamp(max=0) * neg_interval)
                    bias = bias.reshape(
                        bias.shape[0], bias.shape[1], -1).sum(dim=-1)
                    last_A = last_A * torch.logical_not(mask_unstable_grad)
                else:
                    bias = 0
                A = last_A * mask_active
                A_pos = last_A.clamp(min=0) * mask_unstable
                A_neg = last_A.clamp(max=0) * mask_unstable
                return A, A_pos, A_neg, bias
            elif isinstance(last_A, Patches):
                last_A_patches = last_A.patches

                if self.recurjac and self.inputs[0].perturbed:
                    mask_unstable_grad = (
                        (self.inputs[0].lower < 0) * (self.inputs[0].upper > 0))
                    mask_unstable_grad_unfold = _maybe_unfold(
                        mask_unstable_grad, last_A)
                    last_A_unstable = (
                        last_A.to_matrix(mask_unstable_grad.shape)
                        * mask_unstable_grad)
                    bias = (
                        last_A_unstable.clamp(min=0) * pos_interval
                        + last_A_unstable.clamp(max=0) * neg_interval)
                    # FIXME Clean up patches. This implementation does not seem
                    # to support general shapes.
                    assert bias.ndim == 5
                    bias = bias.sum(dim=[-1, -2, -3]).view(-1, 1)
                    last_A_patches = (
                        last_A_patches
                        * torch.logical_not(mask_unstable_grad_unfold))
                else:
                    bias = 0

                # need to unfold mask_active and mask_unstable
                # [batch, 1, in_c, in_H, in_W]
                mask_active_unfold = _maybe_unfold(mask_active, last_A)
                mask_unstable_unfold = _maybe_unfold(mask_unstable, last_A)
                # [spec, batch, 1, in_c, in_H, in_W]

                mask_active_unfold = mask_active_unfold.expand(last_A.shape)
                mask_unstable_unfold = mask_unstable_unfold.expand(last_A.shape)

                A = Patches(
                    last_A_patches * mask_active_unfold,
                    last_A.stride, last_A.padding, last_A.shape,
                    last_A.identity, last_A.unstable_idx, last_A.output_shape)

                A_pos_patches = last_A_patches.clamp(min=0) * mask_unstable_unfold
                A_neg_patches = last_A_patches.clamp(max=0) * mask_unstable_unfold

                A_pos = Patches(
                    A_pos_patches, last_A.stride, last_A.padding, last_A.shape,
                    last_A.identity, last_A.unstable_idx, last_A.output_shape)
                A_neg = Patches(
                    A_neg_patches, last_A.stride, last_A.padding, last_A.shape,
                    last_A.identity, last_A.unstable_idx, last_A.output_shape)

                return A, A_pos, A_neg, bias

        lA, lA_pos, lA_neg, lbias = _bound_oneside(
            last_lA, pos_interval=lower_interval, neg_interval=upper_interval)
        uA, uA_pos, uA_neg, ubias = _bound_oneside(
            last_uA, pos_interval=upper_interval, neg_interval=lower_interval)

        return (
            [(lA, uA), (lA_neg, uA_pos), (lA_pos, uA_neg), (None, None)],
            lbias, ubias)


class BoundConv2dGrad(Bound):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)
        self.stride = attr['stride']
        self.padding = attr['padding']
        self.dilation = attr['dilation']
        self.groups = attr['groups']
        self.output_padding = [
            attr.get('output_padding0', 0),
            attr.get('output_padding1', 0)
        ]
        self.has_bias = len(inputs) == 3
        self.mode = options.get('conv_mode', 'matrix')
        self.patches_start = True

    def forward(self, *x):
        # x[0]: input, x[1]: weight, x[2]: bias if self.has_bias
        return F.conv_transpose2d(
            x[0], x[1], None,
            stride=self.stride, padding=self.padding, dilation=self.dilation,
            groups=self.groups, output_padding=self.output_padding)

    def bound_backward(self, last_lA, last_uA, *x, **kwargs):
        assert not self.is_input_perturbed(1)

        lA_y = uA_y = lA_bias = uA_bias = None
        weight = x[1].lower

        def _bound_oneside(last_A):
            if last_A is None:
                return None, 0

            if isinstance(last_A, torch.Tensor):
                shape = last_A.size()
                next_A = F.conv2d(
                    last_A.reshape(shape[0] * shape[1], *shape[2:]),
                    weight, None, stride=self.stride, padding=self.padding,
                    dilation=self.dilation, groups=self.groups)
                next_A = next_A.view(
                    shape[0], shape[1], *next_A.shape[1:])
                if self.has_bias:
                    sum_bias = (last_A.sum((3, 4)) * x[2].lower).sum(2)
                else:
                    sum_bias = 0
                return next_A, sum_bias
            elif isinstance(last_A, Patches):
                # Here we build and propagate a Patch object with
                # (patches, stride, padding)
                assert self.stride == 1, 'The patches mode only supports stride = 1'
                if last_A.identity == 1:
                    # create a identity patch
                    # [out_dim, batch, out_c, out_h, out_w, in_dim, in_c, in_h, in_w]
                    patch_shape = last_A.shape
                    if last_A.unstable_idx is not None:
                        # FIXME Somehow the usage of unstable_idx seems to have
                        # been changed, and the previous code is no longer working.
                        raise NotImplementedError(
                            'Sparse patches for '
                            'BoundConv2dGrad is not supported yet.')
                        output_shape = last_A.output_shape
                        patches = torch.eye(
                            patch_shape[0]).to(weight)
                        patches = patches.view([
                            patch_shape[0], 1, 1, 1, 1, patch_shape[0], 1, 1])
                        # [out_dim, bsz, out_c, out_h, out_w, out_dim, in_c, in_h, in_w]
                        patches = patches.expand([
                            patch_shape[0], patch_shape[1], patch_shape[2],
                            output_shape[2], output_shape[3],
                            patch_shape[0], 1, 1])
                        patches = patches.transpose(0, 1)
                        patches = patches[
                            :,torch.tensor(list(range(patch_shape[0]))),
                            last_A.unstable_idx[0], last_A.unstable_idx[1],
                            last_A.unstable_idx[2]]
                        patches = patches.transpose(0, 1)
                    else:
                        # out_dim * out_c
                        patches = torch.eye(patch_shape[0]).to(weight)
                        patches = patches.view([
                            patch_shape[0], 1, 1, 1, patch_shape[0], 1, 1])
                        patches = patches.expand(patch_shape)
                else:
                    patches = last_A.patches

                if self.has_bias:
                    # bias is x[2] (lower and upper are the same), and has
                    # shape (c,).
                    # Patches either has
                    # [out_dim, batch, out_c, out_h, out_w, out_dim, c, h, w]
                    # or [unstable_size, batch, out_dim, c, h, w].
                    # sum_bias has shape (out_dim, batch, out_c, out_h, out_w)
                    # or (unstable_size, batch).
                    sum_bias = torch.einsum(
                        'sb...ochw,c->sb...', patches, x[2].lower)
                else:
                    sum_bias = 0

                flattened_patches = patches.reshape(
                    -1, patches.size(-3), patches.size(-2), patches.size(-1))
                # Pad to the full size
                pieces = F.conv2d(
                    flattened_patches, weight, stride=self.stride,
                    padding=weight.shape[2]-1)
                # New patch size:
                # (out_c, batch, out_h, out_w, c, h, w)
                # or (unstable_size, batch, c, h, w).
                pieces = pieces.view(
                    *patches.shape[:-3], pieces.size(-3), pieces.size(-2),
                    pieces.size(-1))

                # (left, right, top, bottom)
                padding = last_A.padding if last_A is not None else (0, 0, 0, 0)
                stride = last_A.stride if last_A is not None else 1

                if isinstance(padding, int):
                    padding = padding + weight.shape[2] - 1
                else:
                    padding = tuple(p + weight.shape[2] - 1 for p in padding)

                return Patches(
                    pieces, stride, padding, pieces.shape,
                    unstable_idx=last_A.unstable_idx,
                    output_shape=last_A.output_shape), sum_bias
            else:
                raise NotImplementedError()

        lA_x, lbias = _bound_oneside(last_lA)
        uA_x, ubias = _bound_oneside(last_uA)
        return [(lA_x, uA_x), (lA_y, uA_y), (lA_bias, uA_bias)], lbias, ubias

    def interval_propagate(self, *v, C=None):
        assert not self.is_input_perturbed(1)

        norm = Interval.get_perturbation(v[0])[0]
        h_L, h_U = v[0]

        weight = v[1][0]
        bias = v[2][0] if self.has_bias else None

        if norm == torch.inf:
            mid = (h_U + h_L) / 2.0
            diff = (h_U - h_L) / 2.0
            weight_abs = weight.abs()
            deviation = F.conv_transpose2d(
                diff, weight_abs, None, stride=self.stride,
                padding=self.padding, dilation=self.dilation,
                groups=self.groups, output_padding=self.output_padding)
        else:
            raise NotImplementedError
        center = F.conv_transpose2d(
            mid, weight, bias, stride=self.stride, padding=self.padding,
            dilation=self.dilation, groups=self.groups,
            output_padding=self.output_padding)
        upper = center + deviation
        lower = center - deviation
        return lower, upper



from .base import *

class BoundDropout(Bound):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)
        if 'ratio' in attr:
            self.ratio = attr['ratio']
            self.dynamic = False
        else:
            self.ratio = None
            self.dynamic = True
        self.clear()

    def clear(self):
        self.mask = None

    def forward(self, *inputs):
        x = inputs[0]
        if not self.training:
            return x
        if self.dynamic:
            # Inputs: data, ratio (optional), training_mode (optional)
            # We assume ratio must exist in the inputs.
            # We ignore training_mode, but will use self.training which can be
            # changed after BoundedModule is built.
            assert (inputs[1].dtype == torch.float32 or
                    inputs[1].dtype == torch.float64)
            self.ratio = inputs[1]
        if self.ratio >= 1:
            raise ValueError('Ratio in dropout should be less than 1')
        self.mask = torch.rand(x.shape, device=self.ratio.device) > self.ratio
        return x * self.mask / (1 - self.ratio)

    def _check_forward(self):
        """ If in the training mode, a forward pass should have been called."""
        if self.training and self.mask is None:
            raise RuntimeError('For a model with dropout in the training mode, '\
                'a clean forward pass must be called before bound computation')

    def bound_backward(self, last_lA, last_uA, *args, **kwargs):
        empty_A = [(None, None)] * (len(args) -1)
        if not self.training:
            return [(last_lA, last_uA), *empty_A], 0, 0
        self._check_forward()
        def _bound_oneside(last_A):
            if last_A is None:
                return None
            return last_A * self.mask / (1 - self.ratio)
        lA = _bound_oneside(last_lA)
        uA = _bound_oneside(last_uA)
        return [(lA, uA), *empty_A], 0, 0

    def bound_forward(self, dim_in, x, *args):
        if not self.training:
            return x
        self._check_forward()
        lw = x.lw * self.mask.unsqueeze(1) / (1 - self.ratio)
        lb = x.lb * self.mask / (1 - self.ratio)
        uw = x.uw * self.mask.unsqueeze(1) / (1 - self.ratio)
        ub = x.ub * self.mask / (1 - self.ratio)
        return LinearBound(lw, lb, uw, ub)

    def interval_propagate(self, *v):
        if not self.training:
            return v[0]
        self._check_forward()
        h_L, h_U = v[0]
        lower = h_L * self.mask / (1 - self.ratio)
        upper = h_U * self.mask / (1 - self.ratio)
        return lower, upper



from torch.nn import DataParallel
from .perturbations import *
from .bounded_tensor import BoundedTensor
from itertools import chain

class BoundDataParallel(DataParallel):
    # https://github.com/huanzhang12/CROWN-IBP/blob/master/bound_layers.py
    # This is a customized DataParallel class for our project
    def __init__(self, *inputs, **kwargs):
        super(BoundDataParallel, self).__init__(*inputs, **kwargs)
        self._replicas = None

    # Overide the forward method
    def forward(self, *inputs, **kwargs):
        disable_multi_gpu = False  # forward by single GPU
        no_replicas = False  # forward by multi GPUs but without replicate
        if "disable_multi_gpu" in kwargs:
            disable_multi_gpu = kwargs["disable_multi_gpu"]
            kwargs.pop("disable_multi_gpu")

        if "no_replicas" in kwargs:
            no_replicas = kwargs["no_replicas"]
            kwargs.pop("no_replicas")

        if not self.device_ids or disable_multi_gpu:
            if kwargs.pop("get_property", False):
                return self.get_property(self, *inputs, **kwargs)
            return self.module(*inputs, **kwargs)

        if kwargs.pop("get_property", False):
            if self._replicas is None:
                assert 0, 'please call IBP/CROWN before get_property'
            if len(self.device_ids) == 1:
                return self.get_property(self.module, **kwargs)
            inputs, kwargs = self.scatter(inputs, kwargs, self.device_ids)
            kwargs = list(kwargs)
            for i in range(len(kwargs)):
                kwargs[i]['model'] = self._replicas[i]
            outputs = self.parallel_apply([self.get_property] * len(kwargs), inputs, kwargs)
            return self.gather(outputs, self.output_device)

        # Only replicate during forward/IBP propagation. Not during interval bounds
        # and CROWN-IBP bounds, since weights have not been updated. This saves 2/3
        # of communication cost.
        if not no_replicas:
            if self._replicas is None:  # first time
                self._replicas = self.replicate(self.module, self.device_ids)
            elif kwargs.get("method_opt", "forward") == "forward":
                self._replicas = self.replicate(self.module, self.device_ids)
            elif kwargs.get("x") is not None and kwargs.get("IBP") is True:  #
                self._replicas = self.replicate(self.module, self.device_ids)
            # Update the input nodes to the ones within each replica respectively
            for bounded_module in self._replicas:
                for node in bounded_module._modules.values():
                    node.inputs = [bounded_module[name] for name in node.input_name]

        for t in chain(self.module.parameters(), self.module.buffers()):
            if t.device != self.src_device_obj:
                raise RuntimeError("module must have its parameters and buffers "
                                   "on device {} (device_ids[0]) but found one of "
                                   "them on device: {}".format(self.src_device_obj, t.device))

        # TODO: can be done in parallel, only support same ptb for all inputs per forward/IBP propagation
        if len(inputs) > 0 and hasattr(inputs[0], 'ptb') and inputs[0].ptb is not None:
            # compute bounds without x
            # inputs_scatter is a normal tensor, we need to assign ptb to it if inputs is a BoundedTensor
            inputs_scatter, kwargs = self.scatter((inputs, inputs[0].ptb.x_L, inputs[0].ptb.x_U), kwargs,
                                                  self.device_ids)
            # inputs_scatter = inputs_scatter[0]
            bounded_inputs = []
            for input_s in inputs_scatter:  # GPU numbers
                # FIXME other perturbations are not supported yet
                assert isinstance(inputs[0].ptb, PerturbationLpNorm)
                ptb = PerturbationLpNorm(norm=inputs[0].ptb.norm, eps=inputs[0].ptb.eps, x_L=input_s[1], x_U=input_s[2])
                input_s = list(input_s[0])
                input_s[0] = BoundedTensor(input_s[0], ptb)
                input_s = tuple(input_s)
                bounded_inputs.append(input_s)

            # bounded_inputs = tuple(bounded_inputs)
        elif kwargs.get("x") is not None and hasattr(kwargs.get("x")[0], 'ptb') and kwargs.get("x")[0].ptb is not None:
            # compute bounds with x
            # kwargs['x'] is a normal tensor, we need to assign ptb to it
            x = kwargs.get("x")[0]
            bounded_inputs = []
            inputs_scatter, kwargs = self.scatter((inputs, x.ptb.x_L, x.ptb.x_U), kwargs, self.device_ids)
            for input_s, kw_s in zip(inputs_scatter, kwargs):  # GPU numbers
                # FIXME other perturbations are not supported yet
                assert isinstance(x.ptb, PerturbationLpNorm)
                ptb = PerturbationLpNorm(norm=x.ptb.norm, eps=x.ptb.eps, x_L=input_s[1], x_U=input_s[2])
                kw_s['x'] = list(kw_s['x'])
                kw_s['x'][0] = BoundedTensor(kw_s['x'][0], ptb)
                kw_s['x'] = (kw_s['x'])
                bounded_inputs.append(tuple(input_s[0], ))
        else:
            # normal forward
            inputs_scatter, kwargs = self.scatter(inputs, kwargs, self.device_ids)
            bounded_inputs = inputs_scatter

        if len(self.device_ids) == 1:
            return self.module(*bounded_inputs[0], **kwargs[0])
        outputs = self.parallel_apply(self._replicas[:len(bounded_inputs)], bounded_inputs, kwargs)
        return self.gather(outputs, self.output_device)

    @staticmethod
    def get_property(model, node_class=None, att_name=None, node_name=None):
        if node_name:
            # Find node by name
            # FIXME If we use `model.named_modules()`, the nodes have the
            # `BoundedModule` type rather than bound nodes.
            for node in model._modules.values():
                if node.name == node_name:
                    return getattr(node, att_name)
        else:
            # Find node by class
            for _, node in model.named_modules():
                # Find the Exp neuron in computational graph
                if isinstance(node, node_class):
                    return getattr(node, att_name)

    def state_dict(self, destination=None, prefix='', keep_vars=False):
        # add 'module.' here before each keys in self.module.state_dict() if needed
        return self.module.state_dict(destination=destination, prefix=prefix, keep_vars=keep_vars)

    def _named_members(self, get_members_fn, prefix='', recurse=True):
        return self.module._named_members(get_members_fn, prefix, recurse)




"""BoundTanh and similar ops."""
import warnings
import torch
from .base import *
from .activation_base import BoundOptimizableActivation


def dtanh(x):
    # to avoid bp error when cosh is too large
    # cosh(25.0)**2 > 1e21
    mask = torch.lt(torch.abs(x), 25.0).to(x.dtype)
    cosh = torch.cosh(mask * x + 1 - mask)
    return mask * (1. / cosh.pow(2))

def dsigmoid(x):
    return torch.sigmoid(x) * (1 - torch.sigmoid(x))

def darctan(x):
    return (x.square() + 1.).reciprocal()


class BoundTanh(BoundOptimizableActivation):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None,
                 activation=('tanh', torch.tanh, dtanh), precompute=True):
        super().__init__(attr, inputs, output_index, options)
        if options is None:
            options = {}
        self.splittable = True
        self.ibp_intermediate = True
        self.activation = activation
        self.activation_name = activation[0]
        self.activation_forward = activation[1]
        self.activation_backward = activation[2]
        if precompute:
            self.precompute_relaxation(*activation)
        # TODO make them configurable when implementing a general nonlinear activation.
        # Neurons whose gap between pre-activation bounds is smaller than this
        # threshold will be masked and don't need branching.
        self.split_min_gap = 1e-2 #1e-4
        # Neurons whose pre-activation bounds don't overlap with this range
        # are considered as stable (with values either 0 or 1) and don't need
        # branching.
        self.split_range = (-10, 10)
        # The initialization will be adjusted if the pre-activation bounds are too loose.
        self.loose_threshold = options.get('tanh', {}).get(
            'loose_threshold', None)

    def opt_init(self):
        super().opt_init()
        self.tp_both_lower_init = {}
        self.tp_both_upper_init = {}

    def _init_opt_parameters_impl(self, size_spec, name_start):
        """Implementation of init_opt_parameters for each start_node."""
        l, u = self.inputs[0].lower, self.inputs[0].upper
        shape = l.shape
        # Alpha dimension is (8, output_shape, batch, *shape) for Tanh.
        alpha = torch.empty(8, size_spec, *shape, device=l.device)
        alpha.data[:4] = ((l + u) / 2)
        alpha.data[4:6] = self.tp_both_lower_init[name_start]
        alpha.data[6:8] = self.tp_both_upper_init[name_start]
        return alpha

    @torch.no_grad()
    def precompute_relaxation(self, name, func, dfunc, x_limit=500):
        """
        This function precomputes the tangent lines that will be used as
        lower/upper bounds for S-shapes functions.
        """
        self.x_limit = x_limit
        self.step_pre = 0.01
        self.num_points_pre = int(self.x_limit / self.step_pre)
        max_iter = 100

        logger.debug('Precomputing relaxation for %s (pre-activation limit: %f)',
                     name, x_limit)

        def check_lower(upper, d):
            """Given two points upper, d (d <= upper), check if the slope at d will be less than f(upper) at upper."""
            k = dfunc(d)
            # Return True if the slope is a lower bound.
            return k * (upper - d) + func(d) <= func(upper)

        def check_upper(lower, d):
            """Given two points lower, d (d >= lower), check if the slope at d will be greater than f(lower) at lower."""
            k = dfunc(d)
            # Return True if the slope is a upper bound.
            return k * (lower - d) + func(d) >= func(lower)

        # Given an upper bound point (>=0), find a line that is guaranteed to be a lower bound of this function.
        upper = self.step_pre * torch.arange(0, self.num_points_pre + 5, device=self.device)
        r = torch.zeros_like(upper)
        # Initial guess, the tangent line is at -1.
        l = -torch.ones_like(upper)
        while True:
            # Check if the tangent line at the guessed point is an lower bound at f(upper).
            checked = check_lower(upper, l).int()
            # If the initial guess is not smaller enough, then double it (-2, -4, etc).
            l = checked * l + (1 - checked) * (l * 2)
            if checked.sum() == l.numel():
                break
        # Now we have starting point at l, its tangent line is guaranteed to be an lower bound at f(upper).
        # We want to further tighten this bound by moving it closer to 0.
        for _ in range(max_iter):
            # Binary search.
            m = (l + r) / 2
            checked = check_lower(upper, m).int()
            l = checked * m + (1 - checked) * l
            r = checked * r + (1 - checked) * m
        # At upper, a line with slope l is guaranteed to lower bound the function.
        self.d_lower = l.clone()

        # Do the same again:
        # Given an lower bound point (<=0), find a line that is guaranteed to be an upper bound of this function.
        lower = -self.step_pre * torch.arange(0, self.num_points_pre + 5, device=self.device)
        l = torch.zeros_like(upper)
        r = torch.ones_like(upper)
        while True:
            checked = check_upper(lower, r).int()
            r = checked * r + (1 - checked) * (r * 2)
            if checked.sum() == l.numel():
                break
        for _ in range(max_iter):
            m = (l + r) / 2
            checked = check_upper(lower, m).int()
            l = (1 - checked) * m + checked * l
            r = (1 - checked) * r + checked * m
        self.d_upper = r.clone()

        logger.debug('Done')

    def forward(self, x):
        return self.activation_forward(x)

    def bound_relax_impl(self, x, func, dfunc):
        lower, upper = x.lower, x.upper
        y_l, y_u = func(lower), func(upper)
        # k_direct is the slope of the line directly connect (lower, func(lower)), (upper, func(upper)).
        k_direct = k = (y_u - y_l) / (upper - lower).clamp(min=1e-8)

        # Fixed bounds that cannot be optimized. self.mask_neg are the masks for neurons with upper bound <= 0.
        # Upper bound for the case of input lower bound <= 0, is always the direct line.
        self.add_linear_relaxation(
            mask=self.mask_neg, type='upper', k=k_direct, x0=lower, y0=y_l)
        # Lower bound for the case of input upper bound >= 0, is always the direct line.
        self.add_linear_relaxation(
            mask=self.mask_pos, type='lower', k=k_direct, x0=lower, y0=y_l)

        # Indices of neurons with input upper bound >=0, whose optimal slope to lower bound the function was pre-computed.
        # Note that for neurons with also input lower bound >=0, they will be masked later.
        index = torch.max(
            torch.zeros(upper.numel(), dtype=torch.long, device=upper.device),
            (upper / self.step_pre).to(torch.long).reshape(-1)
        ) + 1
        if index.max() >= self.d_lower.numel():
            warnings.warn(f'Pre-activation bounds are too loose for {self}')
            # Lookup the lower bound slope from the pre-computed table.
            d_lower = torch.where(
                (index < self.d_lower.numel()).view(lower.shape),
                torch.index_select(
                    self.d_lower, 0, index.clamp(max=self.d_lower.numel() - 1)
                ).view(lower.shape),
                lower,
                # If the pre-activation bounds are too loose, just use IBP.
                # torch.ones_like(index).to(lower) * (-100.)
            ).view(lower.shape)
        else:
            # Lookup the lower bound slope from the pre-computed table.
            d_lower = torch.index_select(
                self.d_lower, 0, index).view(lower.shape)

        # Indices of neurons with lower bound <=0, whose optimal slope to upper
        # bound the function was pre-computed.
        index = torch.max(
            torch.zeros(lower.numel(), dtype=torch.long, device=lower.device),
            (lower / -self.step_pre).to(torch.long).reshape(-1)
        ) + 1
        if index.max() >= self.d_upper.numel():
            warnings.warn(f'Pre-activation bounds are too loose for {self}')
            # Lookup the lower bound slope from the pre-computed table.
            d_upper = torch.where(
                (index < self.d_upper.numel()).view(upper.shape),
                torch.index_select(
                    self.d_upper, 0, index.clamp(max=self.d_upper.numel() - 1)
                ).view(upper.shape),
                upper,
            )
        else:
            d_upper = torch.index_select(
                self.d_upper, 0, index).view(upper.shape)

        if self.opt_stage in ['opt', 'reuse']:
            if not hasattr(self, 'alpha'):
                # Raise an error if alpha is not created.
                self._no_bound_parameters()
            ns = self._start

            # Clipping is done here rather than after `opt.step()` call
            # because it depends on pre-activation bounds
            self.alpha[ns].data[0:2] = torch.max(
                torch.min(self.alpha[ns][0:2], upper), lower)
            self.alpha[ns].data[2:4] = torch.max(
                torch.min(self.alpha[ns][2:4], upper), lower)
            self.alpha[ns].data[4:6] = torch.min(
                self.alpha[ns][4:6], d_lower)
            self.alpha[ns].data[6:8] = torch.max(
                self.alpha[ns][6:8], d_upper)

            # shape [2, out_c, n, c, h, w].
            tp_pos = self.alpha[ns][0:2]  # For upper bound relaxation
            tp_neg = self.alpha[ns][2:4]  # For lower bound relaxation
            tp_both_lower = self.alpha[ns][4:6]
            tp_both_upper = self.alpha[ns][6:8]

            # No need to use tangent line, when the tangent point is at the left
            # side of the preactivation lower bound. Simply connect the two sides.
            mask_direct = torch.logical_and(self.mask_both, k_direct < dfunc(lower))
            self.add_linear_relaxation(
                mask=mask_direct, type='lower', k=k_direct, x0=lower, y0=y_l)
            self.add_linear_relaxation(
                mask=torch.logical_xor(self.mask_both, mask_direct), type='lower',
                k=dfunc(tp_both_lower), x0=tp_both_lower)

            mask_direct = torch.logical_and(self.mask_both, k_direct < dfunc(upper))
            self.add_linear_relaxation(
                mask=mask_direct, type='upper', k=k_direct, x0=lower, y0=y_l)
            self.add_linear_relaxation(
                mask=torch.logical_xor(self.mask_both, mask_direct), type='upper',
                k=dfunc(tp_both_upper), x0=tp_both_upper)

            self.add_linear_relaxation(
                mask=self.mask_neg, type='lower', k=dfunc(tp_neg), x0=tp_neg)
            self.add_linear_relaxation(
                mask=self.mask_pos, type='upper', k=dfunc(tp_pos), x0=tp_pos)
        else:
            if self.opt_stage == 'init':
                # Initialize optimizable slope.
                tp_both_lower_init = d_lower.detach()
                tp_both_upper_init = d_upper.detach()

                if self.loose_threshold is not None:
                    # We will modify d_lower and d_upper inplace.
                    # So make a copy for these two.
                    tp_both_lower_init = tp_both_lower_init.clone()
                    tp_both_upper_init = tp_both_upper_init.clone()
                    # A different initialization if the pre-activation bounds
                    # are too loose
                    loose = torch.logical_or(lower < -self.loose_threshold,
                                            upper > self.loose_threshold)
                    d_lower[loose] = lower[loose]
                    d_upper[loose] = upper[loose]
                    # tp_both_lower_init[loose] = lower[loose]
                    # tp_both_upper_init[loose] = upper[loose]

                ns = self._start
                self.tp_both_lower_init[ns] = tp_both_lower_init
                self.tp_both_upper_init[ns] = tp_both_upper_init

            # Not optimized (vanilla CROWN bound).
            # Use the middle point slope as the lower/upper bound. Not optimized.
            m = (lower + upper) / 2
            y_m = func(m)
            k = dfunc(m)
            # Lower bound is the middle point slope for the case input upper bound <= 0.
            # Note that the upper bound in this case is the direct line between (lower, func(lower)) and (upper, func(upper)).
            self.add_linear_relaxation(mask=self.mask_neg, type='lower', k=k, x0=m, y0=y_m)
            # Upper bound is the middle point slope for the case input lower bound >= 0.
            # Note that the lower bound in this case is the direct line between (lower, func(lower)) and (upper, func(upper)).
            self.add_linear_relaxation(mask=self.mask_pos, type='upper', k=k, x0=m, y0=y_m)

            # Now handle the case where input lower bound <=0 and upper bound >= 0.
            # A tangent line starting at d_lower is guaranteed to be a lower bound given the input upper bound.
            k = dfunc(d_lower)
            # Another possibility is to use the direct line as the lower bound, when this direct line does not intersect with f.
            # This is only valid when the slope at the input lower bound has a slope greater than the direct line.
            mask_direct = torch.logical_and(self.mask_both, k_direct < dfunc(lower))
            self.add_linear_relaxation(mask=mask_direct, type='lower', k=k_direct, x0=lower, y0=y_l)
            # Otherwise we do not use the direct line, we use the d_lower slope.
            self.add_linear_relaxation(
                mask=torch.logical_xor(self.mask_both, mask_direct),
                type='lower', k=k, x0=d_lower)

            # Do the same for the upper bound side when input lower bound <=0 and upper bound >= 0.
            k = dfunc(d_upper)
            mask_direct = torch.logical_and(self.mask_both, k_direct < dfunc(upper))
            self.add_linear_relaxation(
                mask=mask_direct, type='upper', k=k_direct, x0=lower, y0=y_l)
            self.add_linear_relaxation(
                mask=torch.logical_xor(self.mask_both, mask_direct),
                type='upper', k=k, x0=d_upper)

    def bound_relax(self, x, init=False, dim_opt=None):
        if init:
            self.init_linear_relaxation(x, dim_opt)
        self.bound_relax_impl(
            x, self.activation_forward, self.activation_backward)

    def get_split_mask(self, lower, upper, input_index):
        assert input_index == 0
        return torch.logical_and(
            upper - lower >= self.split_min_gap,
            torch.logical_or(upper >= self.split_range[0],
                             lower <= self.split_range[1])
        )

    def get_split_point(self):
        return None
    
    
class BoundSigmoid(BoundTanh):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options,
                         activation=('sigmoid', torch.sigmoid, dsigmoid))


class BoundAtan(BoundTanh):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options,
                         activation=('arctan', torch.arctan, darctan))
        self.split_range = (-torch.inf, torch.inf)


class BoundTan(BoundAtan):
    """
    The implementation of BoundTan is based on the S-shaped BoundAtan. We use the bounds from its
    inverse function and directly convert the bounds of the inverse function to bounds of the original
    function. This trick allows us to quickly implement bounds on inverse functions.
    """

    def forward(self, x):
        return torch.tan(x)

    def _check_bounds(self, lower, upper):
        # Lower and upper bounds must be within the same [-½π, ½π] region.
        lower_periods = torch.floor((lower + 0.5 * torch.pi) / torch.pi)
        upper_periods = torch.floor((upper + 0.5 * torch.pi) / torch.pi)
        if not torch.allclose(lower_periods, upper_periods):
            print('Tan preactivation lower bounds:\n', lower)
            print('Tan preactivation upper bounds:\n', upper)
            raise ValueError("BoundTan received pre-activation bounds that produce infinity. "
                    "The preactivation bounds are too loose. Try to reduce perturbation region.")
        # Return the period number for each neuron.
        # Period is 0 => bounds are within [-½π, ½π],
        # Period is 1 => bounds are within [-½π + π, ½π + π]
        # Period is -1 => bounds are within [-½π - π, ½π - π]
        return lower_periods

    def _init_masks(self, x):
        # The masks now must consider the periodicity.
        lower = torch.remainder(x.lower + 0.5 * torch.pi, torch.pi) - 0.5 * torch.pi
        upper = torch.remainder(x.upper + 0.5 * torch.pi, torch.pi) - 0.5 * torch.pi
        self.mask_pos = lower >= 0
        self.mask_neg = upper <= 0
        self.mask_both = torch.logical_not(torch.logical_or(self.mask_pos, self.mask_neg))

    def interval_propagate(self, *v):
        # We need to check if the input lower and upper bounds are within the same period.
        # Otherwise the bounds become infinity.
        concrete_lower, concrete_upper = v[0][0], v[0][1]
        self._check_bounds(concrete_lower, concrete_upper)
        return super().interval_propagate(*v)

    def bound_relax(self, x, init=False, dim_opt=None):
        if init:
            self.init_linear_relaxation(x, dim_opt)
        periods = self._check_bounds(x.lower, x.upper)
        periods = torch.pi * periods
        # Create a fake x with inversed lower and upper.
        inverse_x = lambda: None
        inverse_x.lower = torch.tan(x.lower)
        inverse_x.upper = torch.tan(x.upper)
        super().bound_relax(inverse_x, init=init)
        # Lower slope, lower bias, upper slope and upper bias are saved to
        # self.lw, self.lb, self.uw, self.ub. We need to reverse them.
        # E.g., y = self.lw * x + self.lb, now becomes x = 1./self.lw * y - self.lb / self.lw
        # Additionally, we need to add the missing ½π periods.
        new_upper_slope = 1. / self.lw
        new_upper_bias = - self.lb / self.lw - periods / self.lw
        new_lower_slope = 1. / self.uw
        new_lower_bias = - self.ub / self.uw - periods / self.uw
        self.lw = new_lower_slope
        self.lb = new_lower_bias
        self.uw = new_upper_slope
        self.ub = new_upper_bias




# import torchvision.transforms as transforms
# from PIL import Image
import matplotlib.pyplot as plt
from pathlib import Path
import torch

from test import extract_instance
from heuristic.ig import *

def normalize(img):
    A = img.clone()
    A -= A.min(1, keepdim=True)[0]
    A /= A.max(1, keepdim=True)[0]
    return A

def print_selected_indices(upper, lower, indices, name=None):
    print(f'[{name}] Selected indices:', indices)
    for ind in indices:
        print(f'\t + lower[{ind}]={lower.flatten()[ind]:.03f} \t upper[{ind}]={upper.flatten()[ind]:.03f} \t range[{ind}]={(upper.flatten()[ind]-lower.flatten()[ind]):.03f}')
    print()

if __name__ == "__main__":
    # path = 'example/doberman.png'

    # net_path = 'example/cifar10_2_255_simplified.onnx'
    # net_path = '../benchmark/vnncomp23/acasxu/onnx/ACASXU_run2a_1_1_batch_2000.onnx'
    net_path = '../benchmark/vnncomp23/yolo/onnx/TinyYOLO.onnx'
    
    # vnnlib_path = Path('example/cifar10_spec_idx_4_eps_0.00784_n1.vnnlib')
    # vnnlib_path = Path('../benchmark/vnncomp23/acasxu/vnnlib/prop_3.vnnlib')
    vnnlib_path = Path('../benchmark/vnncomp23/yolo/vnnlib/TinyYOLO_prop_000246_eps_1_255.vnnlib')
    
    device = 'cpu'
    
    model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
    model.to(device)
    
    # transform = transforms.Compose([
    #     transforms.PILToTensor()
    # ])
    objective = objectives.pop(1)
    
    input_lowers = objective.lower_bounds.view(input_shape)
    input_uppers = objective.upper_bounds.view(input_shape)
    print(input_lowers.shape)
    
    random_input = (input_uppers - input_lowers) * torch.rand(input_shape, device=device) + input_lowers
    assert torch.all(random_input <= input_uppers)
    assert torch.all(random_input >= input_lowers)
    
    ig = IntegratedGradient(model, device=device)
    attr, label = ig.attribute(random_input)
    
    
    indices = (input_uppers - input_lowers)[0].flatten().topk(5, 0).indices
    print_selected_indices(input_uppers, input_lowers, indices, name='diff')
    
    indices = attr[0].flatten().topk(5, 0).indices
    print_selected_indices(input_uppers, input_lowers, indices, name='ig')
    
    scaled_attr = scale_saliency_map_np(attr[0].detach().cpu().numpy(), dim=0)
    print(scaled_attr.shape)
    indices = torch.tensor(scaled_attr).flatten().topk(5, 0).indices
    print_selected_indices(input_uppers, input_lowers, indices, name='scaled ig')

    if 1:
        fig, axes = plt.subplots(1, 3)
        axes[0].imshow(random_input.permute(0, 2, 3, 1)[0].detach().cpu().numpy())
        axes[0].set_xticks([])
        axes[0].set_yticks([])
        axes[0].set_title('Image')
        
        norm_attr = normalize(attr)
        axes[1].imshow(norm_attr.permute(0, 2, 3, 1)[0].detach().cpu().numpy(), cmap='RdBu_r')
        axes[1].set_xticks([])
        axes[1].set_yticks([])
        axes[1].set_title('IG')
        
        
        axes[2].imshow(scaled_attr, cmap='RdBu_r')
        axes[2].set_xticks([])
        axes[2].set_yticks([])
        axes[2].set_title('Scaled IG')
        
        fig.tight_layout()
        plt.savefig('example/output.png')
    
    # img = Image.open(path).convert('RGB')
    # img_norm = transform(img)
    # print(img_norm.shape)
  
    # ig_attr, label = ig.attribute(img_norm)


from verifier.interactive_verifier import InteractiveVerifier
from test import reset_settings, extract_instance
# from util.misc.logger import logger
# import logging

if __name__ == "__main__":
    
    ############## Preprocess ##############
    
    # logger.setLevel(logging.INFO)
    reset_settings()
    
    net_path = 'example/onnx/mnistfc-medium-net-151.onnx'
    vnnlib_path = 'example/vnnlib/prop_2_0.03.vnnlib'
    device = 'cpu'
    batch = 100
    print(f'\n\nRunning test with {net_path=} {vnnlib_path=}')
    
    model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
    model.to(device)
    
    env = InteractiveVerifier(
        net=model,
        input_shape=input_shape,
        batch=batch,
        device=device,
    )
    
    objectives, _ = env._preprocess(objectives, force_split=None)
    objective = objectives.pop(1)
    
    ############## Verify ##############
    

    done = env.init(
        objective=objective, 
        reference_bounds=None,
        preconditions={}, 
    )
    
    while not done:
        action, state = env.decide()
        obs, reward, done, info = env.step(action)
        print(len(env.domains_list), info)
    


import copy
from typing import List
import numpy as np
import warnings
from collections import OrderedDict, deque

import torch
from torch.nn import Parameter

from .bound_op_map import bound_op_map
from .bound_ops import *
from .bounded_tensor import BoundedTensor, BoundedParameter
from .parse_graph import parse_module
from .perturbations import *
from .utils import *
from .patches import Patches
from .optimized_bounds import default_optimize_bound_args


warnings.simplefilter('once')


class BoundedModule(nn.Module):
    """Bounded module with support for automatically computing bounds.

    Args:
        model (nn.Module): The original model to be wrapped by BoundedModule.

        global_input (tuple): A dummy input to the original model. The shape of
        the dummy input should be consistent with the actual input to the model
        except for the batch dimension.

        bound_opts (dict): Options for bounds. See
        `Bound Options <bound_opts.html>`_.

        device (str or torch.device): Device of the bounded module.
        If 'auto', the device will be automatically inferred from the device of
        parameters in the original model or the dummy input.

        custom_ops (dict): A dictionary of custom operators.
        The dictionary maps operator names to their corresponding bound classes
        (subclasses of `Bound`).

    """
    def __init__(self, model, global_input, bound_opts=None,
                device='auto', verbose=False, custom_ops=None):
        super().__init__()
        if isinstance(model, BoundedModule):
            for key in model.__dict__.keys():
                setattr(self, key, getattr(model, key))
            return

        self.global_input = global_input
        self.ori_training = model.training
        self.check_incompatible_nodes(model)

        if bound_opts is None:
            bound_opts = {}
        # Default options.
        default_bound_opts = {
            'conv_mode': 'patches',
            'sparse_intermediate_bounds': True,
            'sparse_conv_intermediate_bounds': True,
            'sparse_intermediate_bounds_with_ibp': True,
            'sparse_features_alpha': True,
            'sparse_spec_alpha': True,
            'deterministic': False,
            'tanh': {'loose_threshold': None},
            'fixed_reducemax_index': True,
            'matmul': {'share_alphas': False},
            'minimum_sparsity': 0.9,
            'enable_opt_interm_bounds': False,
            'crown_batch_size': int(1e9),
            'forward_refinement': False,
            'dynamic_forward': False,
            'forward_max_dim': int(1e9),
            # Do not share alpha for conv layers.
            'use_full_conv_alpha': True,
            'max_crown_size': int(1e9),
            'disable_optimization': [],
            'disabled_optimization': [],
            # Threshold for number of unstable neurons for each layer to disable use_full_conv_alpha.
            'use_full_conv_alpha_thresh': 512,
            'verbosity': 1 if verbose else 0,
            'optimize_graph': {'optimizer': None},
        }
        default_bound_opts.update(bound_opts)
        self.bound_opts = default_bound_opts
        optimize_bound_args = default_optimize_bound_args
        optimize_bound_args.update(self.bound_opts.get('optimize_bound_args', {}))
        self.bound_opts.update({'optimize_bound_args': optimize_bound_args})

        self.verbose = verbose
        self.custom_ops = custom_ops if custom_ops is not None else {}
        if device == 'auto':
            try:
                self.device = next(model.parameters()).device
            except StopIteration:
                # Model has no parameters. We use the device of input tensor.
                self.device = global_input.device
        else:
            self.device = device
        self.conv_mode = self.bound_opts.get('conv_mode', 'patches')
        # Cached IBP results which may be reused
        self.ibp_lower, self.ibp_upper = None, None

        self.optimizable_activations = []
        self.relus = []  # save relu layers for convenience
        self.layers_with_constraint = []

        state_dict_copy = copy.deepcopy(model.state_dict())
        object.__setattr__(self, 'ori_state_dict', state_dict_copy)
        model.to(self.device)
        self.final_shape = model(
            *unpack_inputs(global_input, device=self.device)).shape
        self.bound_opts.update({'final_shape': self.final_shape})
        self._convert(model, global_input)
        self._mark_perturbed_nodes()
        self._optimize_graph()
        self._expand_jacobian()

        self.next_split_hint = []  # Split hints, used in beta optimization.
        # Beta values for all intermediate bounds.
        # Set to None (not used) by default.
        self.best_intermediate_betas = None
        # Initialization value for intermediate betas.
        self.init_intermediate_betas = None

    def nodes(self) -> List[Bound]:
        return self._modules.values()

    def get_enabled_opt_act(self):
        # Optimizable activations that are actually used and perturbed
        return [
            n for n in self.optimizable_activations
            if n.used and n.perturbed and not getattr(n, 'is_linear_op', False)
        ]

    def get_optimizable_activations(self):
        for node in self.nodes():
            if (isinstance(node, BoundOptimizableActivation)
                    and node.optimizable
                    and len(getattr(node, 'requires_input_bounds', [])) > 0
                    and node not in self.optimizable_activations):
                disabled = False
                for item in self.bound_opts.get('disable_optimization', []):
                    if item.lower() in str(type(node)).lower():
                        disabled = True
                if disabled:
                    logging.info('Disabled optimization for %s', node)
                    continue
                if node not in self.optimizable_activations:
                    self.optimizable_activations.append(node)
            if isinstance(node, BoundRelu) and node not in self.relus:
                self.relus.append(node)

    def get_perturbed_optimizable_activations(self):
        return [n for n in self.optimizable_activations if n.perturbed]

    def get_splittable_activations(self):
        """Activation functions that can be split during branch and bound."""
        return [n for n in self.nodes() if n.perturbed and n.splittable]

    def get_layers_requiring_bounds(self):
        """Layer names whose intermediate layer bounds are required."""
        intermediate_layers = []
        for node in self.nodes():
            if not node.used or not node.perturbed:
                continue
            for i in getattr(node, 'requires_input_bounds', []):
                input_node = node.inputs[i]
                if (input_node not in intermediate_layers
                        and input_node.perturbed):
                    # If not perturbed, it may not have the batch dimension.
                    # So we do not include it, and it is unnecessary.
                    intermediate_layers.append(input_node)
            if (node.name in self.layers_with_constraint
                    and node not in intermediate_layers):
                intermediate_layers.append(node)
        return intermediate_layers

    def check_incompatible_nodes(self, model):
        """Check whether the model has incompatible nodes that the conversion
        may be inaccurate"""
        node_types = [type(m) for m in list(model.modules())]

        if (torch.nn.Dropout in node_types
                and torch.nn.BatchNorm1d in node_types
                and self.global_input.shape[0] == 1):
            print('We cannot support torch.nn.Dropout and torch.nn.BatchNorm1d '
                  'at the same time!')
            print('Suggest to use another dummy input which has batch size '
                  'larger than 1 and set model to train() mode.')
            return

        if not self.ori_training and torch.nn.Dropout in node_types:
            print('Dropout operation CANNOT be parsed during conversion when '
                  'the model is in eval() mode!')
            print('Set model to train() mode!')
            self.ori_training = True

        if self.ori_training and torch.nn.BatchNorm1d in node_types:
            print('BatchNorm1d may raise error during conversion when the model'
                  ' is in train() mode!')
            print('Set model to eval() mode!')
            self.ori_training = False

    def non_deter_wrapper(self, op, *args, **kwargs):
        """Some operations are non-deterministic and deterministic mode will
        fail. So we temporary disable it."""
        if self.bound_opts.get('deterministic', False):
            torch.use_deterministic_algorithms(False)
        ret = op(*args, **kwargs)
        if self.bound_opts.get('deterministic', False):
            torch.use_deterministic_algorithms(True)
        return ret

    def non_deter_scatter_add(self, *args, **kwargs):
        return self.non_deter_wrapper(torch.scatter_add, *args, **kwargs)

    def non_deter_index_select(self, *args, **kwargs):
        return self.non_deter_wrapper(torch.index_select, *args, **kwargs)

    def set_bound_opts(self, new_opts):
        for k, v in new_opts.items():
            # assert v is not dict, 'only support change optimize_bound_args'
            if type(v) == dict:
                self.bound_opts[k].update(v)
            else:
                self.bound_opts[k] = v

    @staticmethod
    def _get_A_norm(A):
        if not isinstance(A, (list, tuple)):
            A = (A, )
        norms = []
        for aa in A:
            if aa is not None:
                if isinstance(aa, Patches):
                    aa = aa.patches
                norms.append(aa.abs().sum().item())
            else:
                norms.append(None)
        return norms

    def __call__(self, *input, **kwargs):
        if 'method_opt' in kwargs:
            opt = kwargs['method_opt']
            kwargs.pop('method_opt')
        else:
            opt = 'forward'
        for kwarg in [
            'disable_multi_gpu', 'no_replicas', 'get_property',
            'node_class', 'att_name']:
            if kwarg in kwargs:
                kwargs.pop(kwarg)
        if opt == 'compute_bounds':
            return self.compute_bounds(**kwargs)
        else:
            return self.forward(*input, **kwargs)

    def register_parameter(self, name, param):
        r"""Adds a parameter to the module.

        The parameter can be accessed as an attribute using given name.

        Args:
            name (string): name of the parameter. The parameter can be accessed
                from this module using the given name
            param (Parameter): parameter to be added to the module.
        """
        if '_parameters' not in self.__dict__:
            raise AttributeError(
                'cannot assign parameter before Module.__init__() call')
        elif not isinstance(name, str):
            raise TypeError('parameter name should be a string. '
                            f'Got {torch.typename(name)}')
        elif name == '':
            raise KeyError('parameter name can\'t be empty string')
        elif hasattr(self, name) and name not in self._parameters:
            raise KeyError(f'attribute "{name}" already exists')

        if param is None:
            self._parameters[name] = None
        elif not isinstance(param, Parameter):
            raise TypeError(
                f'cannot assign "{torch.typename(param)}" object to '
                f'parameter "{name}" '
                '(torch.nn.Parameter or None required)')
        elif param.grad_fn:
            raise ValueError(
                f'Cannot assign non-leaf Tensor to parameter "{name}". Model '
                'parameters must be created explicitly. To express "{name}" '
                'as a function of another Tensor, compute the value in '
                'the forward() method.')
        else:
            self._parameters[name] = param

    def load_state_dict(self, state_dict, strict=False):
        new_dict = OrderedDict()
        # translate name to ori_name
        for k, v in state_dict.items():
            if k in self.node_name_map:
                new_dict[self.node_name_map[k]] = v
        return super().load_state_dict(new_dict, strict=strict)

    def _named_members(self, get_members_fn, prefix='', recurse=True, **kwargs):  # pylint: disable=unused-argument
        r"""Helper method for yielding various names + members of modules."""
        memo = set()
        modules = self.named_modules(prefix=prefix) if recurse else [
                                     (prefix, self)]
        # TODO: support the "remove_duplicate" argument, new in pytorch 2.0.
        for module_prefix, module in modules:
            members = get_members_fn(module)
            for k, v in members:
                if v is None or v in memo:
                    continue
                memo.add(v)
                name = module_prefix + ('.' if module_prefix else '') + k
                # translate name to ori_name
                if name in self.node_name_map:
                    name = self.node_name_map[name]
                yield name, v

    def train(self, mode=True):
        super().train(mode)
        for node in self.nodes():
            node.train(mode=mode)

    def eval(self):
        super().eval()
        for node in self.nodes():
            node.eval()

    def to(self, *args, **kwargs):
        # Moves and/or casts some attributes except pytorch will do by default.
        for node in self.nodes():
            for attr in ['lower', 'upper', 'forward_value', 'd', 'lA',]:
                if hasattr(node, attr):
                    this_attr = getattr(node, attr)
                    if isinstance(this_attr, torch.Tensor):
                        this_attr = this_attr.to(*args, **kwargs)
                        setattr(node, attr, this_attr)

            if hasattr(node, 'interval'):
                # construct new interval
                this_attr = getattr(node, 'interval')
                setattr(node, 'interval', (this_attr[0].to(
                    *args, **kwargs), this_attr[1].to(*args, **kwargs)))

        return super().to(*args, **kwargs)

    def __getitem__(self, name):
        module = self._modules[name]
        # We never create modules that are None, the assert fixes type hints
        assert module is not None
        return module

    def roots(self):
        return [self[name] for name in self.root_names]

    def final_node(self):
        return self[self.final_name]

    def get_forward_value(self, node):
        """ Recursively get `forward_value` for `node` and its parent nodes"""
        if getattr(node, 'forward_value', None) is not None:
            return node.forward_value
        inputs = [self.get_forward_value(inp) for inp in node.inputs]
        for inp in node.inputs:
            node.from_input = node.from_input or inp.from_input
        node.input_shape = inputs[0].shape if len(inputs) > 0 else None
        fv = node.forward(*inputs)
        if isinstance(fv, (torch.Size, tuple)):
            fv = torch.tensor(fv, device=self.device)
        node.forward_value = fv
        node.output_shape = fv.shape
        # In most cases, the batch dimension is just the first dimension
        # if the node depends on input. Otherwise if the node doesn't
        # depend on input, there is no batch dimension.
        node.batch_dim = 0 if node.from_input else -1
        # Unperturbed node but it is not a root node.
        # Save forward_value to value. (Can be used in forward bounds.)
        if not node.from_input and len(node.inputs) > 0:
            node.value = node.forward_value
        return fv

    def forward(self, *x, final_node_name=None, clear_forward_only=False,
            reset_perturbed_nodes=True):
        r"""Standard forward computation for the network.

        Args:
            x (tuple or None): Input to the model.

            final_node_name (str, optional): The name of the final node in the
            model. The value on the corresponding node will be returned.

            clear_forward_only (bool, default `False`): Whether only standard
            forward values stored on the nodes should be cleared. If `True`,
            only standard forward values stored on the nodes will be cleared.
            Otherwise, bound information on the nodes will also be cleared.

            reset_perturbed_nodes (bool, default `True`): Mark all perturbed
            nodes with input perturbations. When set to `True`, it may
            accidentally clear all .perturbed properties for intermediate
            nodes.

        Returns:
            output: The output of the model, or if `final_node_name` is not
            `None`, return the value on the corresponding node instead.
        """
        self.set_input(*x, clear_forward_only=clear_forward_only,
                reset_perturbed_nodes=reset_perturbed_nodes)
        if final_node_name:
            return self.get_forward_value(self[final_node_name])
        else:
            return fill_template(
                deque([self.get_forward_value(self[n])
                       for n in self.output_name]),
                self.output_template)

    def _mark_perturbed_nodes(self):
        """Mark the graph nodes and determine which nodes need perturbation."""
        degree_in = {}
        queue = deque()
        # Initially the queue contains all "root" nodes.
        for key in self._modules.keys():
            l = self[key]
            degree_in[l.name] = len(l.inputs)
            if degree_in[l.name] == 0:
                queue.append(l)  # in_degree ==0 -> root node

        while len(queue) > 0:
            node = queue.popleft()
            # Obtain all output node, and add the output nodes to the queue if
            # all its input nodes have been visited.
            # The initial "perturbed" property is set in BoundInput or
            # BoundParams object, depending on ptb.
            for name_next in node.output_name:
                node_next = self[name_next]
                if not node_next.never_perturbed:
                    # The next node is perturbed if it is already perturbed,
                    # or this node is perturbed.
                    node_next.perturbed = node_next.perturbed or node.perturbed
                degree_in[name_next] -= 1
                # all inputs of this node have been visited,
                # now put it in queue.
                if degree_in[name_next] == 0:
                    queue.append(node_next)
            node.update_requires_input_bounds()

        self.get_optimizable_activations()
        self.splittable_activations = self.get_splittable_activations()
        self.perturbed_optimizable_activations = (
            self.get_perturbed_optimizable_activations())
        return

    def _clear_and_set_new(self, interm_bounds, clear_forward_only=False,
                           reset_perturbed_nodes=True):
        for l in self.nodes():
            if hasattr(l, 'linear'):
                if isinstance(l.linear, tuple):
                    for item in l.linear:
                        del item
                delattr(l, 'linear')

            if hasattr(l, 'patch_size'):
                l.patch_size = {}

            if clear_forward_only:
                if hasattr(l, 'forward_value'):
                    delattr(l, 'forward_value')
            else:
                for attr in ['lower', 'upper', 'interval', 'forward_value', 'd', 'lA', 'lower_d']:
                    if hasattr(l, attr):
                        delattr(l, attr)

            for attr in ['zero_backward_coeffs_l', 'zero_backward_coeffs_u', 'zero_lA_mtx', 'zero_uA_mtx']:
                setattr(l, attr, False)
            # Given an interval here to make IBP/CROWN start from this node
            if interm_bounds is not None and l.name in interm_bounds.keys():
                l.interval = tuple(interm_bounds[l.name][:2])
                l.lower = interm_bounds[l.name][0]
                l.upper = interm_bounds[l.name][1]
                if l.lower is not None:
                    l.lower = l.lower.detach().requires_grad_(False)
                if l.upper is not None:
                    l.upper = l.upper.detach().requires_grad_(False)
            # Mark all nodes as non-perturbed except for weights.
            if reset_perturbed_nodes:
                if not hasattr(l, 'perturbation') or l.perturbation is None:
                    l.perturbed = False

            # Clear operator-specific attributes
            l.clear()

    def set_input(self, *x, interm_bounds=None,
                  clear_forward_only=False, reset_perturbed_nodes=True):
        self._clear_and_set_new(
            interm_bounds=interm_bounds,
            clear_forward_only=clear_forward_only,
            reset_perturbed_nodes=reset_perturbed_nodes)
        inputs_unpacked = unpack_inputs(x)
        for name, index in zip(self.input_name, self.input_index):
            if index is None:
                continue
            node = self[name]
            node.value = inputs_unpacked[index]
            if isinstance(node.value, (BoundedTensor, BoundedParameter)):
                node.perturbation = node.value.ptb
            else:
                node.perturbation = None
        # Mark all perturbed nodes.
        if reset_perturbed_nodes:
            self._mark_perturbed_nodes()

    def _get_node_input(self, nodesOP, nodesIn, node):
        ret = []
        for i in range(len(node.inputs)):
            for op in nodesOP:
                if op.name == node.inputs[i]:
                    ret.append(op.bound_node)
                    break
            if len(ret) == i + 1:
                continue
            for io in nodesIn:
                if io.name == node.inputs[i]:
                    ret.append(io.bound_node)
                    break
            if len(ret) <= i:
                raise ValueError(f'cannot find inputs of node: {node.name}')
        return ret

    def _to(self, obj, dest, inplace=False):
        """ Move all tensors in the object to a specified dest
        (device or dtype). The inplace=True option is available for dict."""
        if obj is None:
            return obj
        elif isinstance(obj, torch.Tensor):
            return obj.to(dest)
        elif isinstance(obj, Patches):
            return obj.patches.to(dest)
        elif isinstance(obj, tuple):
            return tuple([self._to(item, dest) for item in obj])
        elif isinstance(obj, list):
            return list([self._to(item, dest) for item in obj])
        elif isinstance(obj, dict):
            if inplace:
                for k, v in obj.items():
                    obj[k] = self._to(v, dest, inplace=True)
                return obj
            else:
                return {k: self._to(v, dest) for k, v in obj.items()}
        else:
            raise NotImplementedError(type(obj))

    def _convert_nodes(self, model, global_input):
        r"""
        Returns:
            nodesOP (list): List of operator nodes
            nodesIn (list): List of input nodes
            nodesOut (list): List of output nodes
            template (object): Template to specify the output format
        """
        global_input_cpu = self._to(global_input, 'cpu')
        if self.ori_training:
            model.train()
        else:
            model.eval()
        model.to('cpu')
        nodesOP, nodesIn, nodesOut, template = parse_module(
            model, global_input_cpu)
        model.to(self.device)
        for i in range(0, len(nodesIn)):
            if nodesIn[i].param is not None:
                nodesIn[i] = nodesIn[i]._replace(
                    param=nodesIn[i].param.to(self.device))
        global_input_unpacked = unpack_inputs(global_input)

        # Convert input nodes and parameters.
        for i, n in enumerate(nodesIn):
            if n.input_index is not None:
                nodesIn[i] = nodesIn[i]._replace(bound_node=BoundInput(
                    ori_name=nodesIn[i].ori_name,
                    value=global_input_unpacked[nodesIn[i].input_index],
                    perturbation=nodesIn[i].perturbation,
                    input_index=n.input_index))
            else:
                bound_class = BoundParams if isinstance(
                    nodesIn[i].param, nn.Parameter) else BoundBuffers
                nodesIn[i] = nodesIn[i]._replace(bound_node=bound_class(
                    ori_name=nodesIn[i].ori_name, value=nodesIn[i].param,
                    perturbation=nodesIn[i].perturbation, options=self.bound_opts))

        unsupported_ops = []

        # Convert other operation nodes.
        for n in range(len(nodesOP)):
            attr = nodesOP[n].attr
            inputs = self._get_node_input(nodesOP, nodesIn, nodesOP[n])
            try:
                if nodesOP[n].op in self.custom_ops:
                    op = self.custom_ops[nodesOP[n].op]
                elif nodesOP[n].op in bound_op_map:
                    op = bound_op_map[nodesOP[n].op]
                elif nodesOP[n].op.startswith('aten::ATen'):
                    op = globals()[f'BoundATen{attr["operator"].capitalize()}']
                elif nodesOP[n].op.startswith('onnx::'):
                    op = globals()[f'Bound{nodesOP[n].op[6:]}']
                else:
                    raise KeyError
            except (NameError, KeyError):
                unsupported_ops.append(nodesOP[n])
                logger.error('The node has an unsupported operation: %s',
                             nodesOP[n])
                continue
            attr['device'] = self.device

            # FIXME generalize
            if (nodesOP[n].op == 'onnx::BatchNormalization'
                    or getattr(op, 'TRAINING_FLAG', False)):
                # BatchNormalization node needs model.training flag to set
                # running mean and vars set training=False to avoid wrongly
                # updating running mean/vars during bound wrapper
                nodesOP[n] = nodesOP[n]._replace(bound_node=op(
                    attr, inputs, nodesOP[n].output_index, self.bound_opts,
                    False))
            else:
                nodesOP[n] = nodesOP[n]._replace(bound_node=op(
                    attr, inputs, nodesOP[n].output_index, self.bound_opts))

        if unsupported_ops:
            logger.error('Unsupported operations:')
            for n in unsupported_ops:
                logger.error(f'Name: {n.op}, Attr: {n.attr}')
            raise NotImplementedError('There are unsupported operations')

        for node in nodesIn + nodesOP:
            node.bound_node.name = node.name

        nodes_dict = {}
        for node in nodesOP + nodesIn:
            nodes_dict[node.name] = node.bound_node
        nodesOP = [n.bound_node for n in nodesOP]
        nodesIn = [n.bound_node for n in nodesIn]
        nodesOut = [nodes_dict[n] for n in nodesOut]

        return nodesOP, nodesIn, nodesOut, template

    def _build_graph(self, nodesOP, nodesIn, nodesOut, template):
        # We were assuming that the original model had only one output node.
        # When there are multiple output nodes, this seems to be the first
        # output element. In this case, we are assuming that we aim to compute
        # the bounds for the first output element by default.
        self.final_name = nodesOut[0].name
        self.input_name, self.input_index, self.root_names = [], [], []
        self.output_name = [n.name for n in nodesOut]
        self.output_template = template
        for node in nodesIn:
            self.add_input_node(node, index=node.input_index)
        self.add_nodes(nodesOP)
        if self.conv_mode == 'patches':
            self.root_names: List[str] = [node.name for node in nodesIn]

    def rename_nodes(self, nodesOP, nodesIn, rename_dict):
        def rename(node):
            node.name = rename_dict[node.name]
            return node
        for i in range(len(nodesOP)):
            nodesOP[i] = rename(nodesOP[i])
        for i in range(len(nodesIn)):
            nodesIn[i] = rename(nodesIn[i])

    def _split_complex(self, nodesOP, nodesIn):
        finished = True
        for n in range(len(nodesOP)):
            if hasattr(nodesOP[n], 'complex') and nodesOP[n].complex:
                finished = False
                _nodesOP, _nodesIn, _nodesOut, _ = self._convert_nodes(
                    nodesOP[n].model, nodesOP[n].input)
                # assuming each supported complex operation only has one output
                assert len(_nodesOut) == 1

                name_base = nodesOP[n].name + '/split'
                rename_dict = {}
                for node in _nodesOP + _nodesIn:
                    rename_dict[node.name] = name_base + node.name
                num_inputs = len(nodesOP[n].inputs)
                for i in range(num_inputs):
                    rename_dict[_nodesIn[i].name] = nodesOP[n].input_name[i]
                rename_dict[_nodesOP[-1].name] = nodesOP[n].name

                self.rename_nodes(_nodesOP, _nodesIn, rename_dict)

                output_name = _nodesOP[-1].name
                # Any input node of some node within the complex node should be
                # replaced with the complex node's corresponding input node.
                for node in _nodesOP:
                    for i in range(len(node.inputs)):
                        if node.input_name[i] in nodesOP[n].input_name:
                            index = nodesOP[n].input_name.index(
                                node.input_name[i])
                            node.inputs[i] = nodesOP[n].inputs[index]
                # For any output node of this complex node,
                # modify its input node.
                for node in nodesOP:
                    if output_name in node.input_name:
                        index = node.input_name.index(output_name)
                        node.inputs[index] = _nodesOP[-1]

                nodesOP = nodesOP[:n] + _nodesOP + nodesOP[(n + 1):]
                nodesIn = nodesIn + _nodesIn[num_inputs:]

                break

        return nodesOP, nodesIn, finished

    def _get_node_name_map(self):
        """Build a dict with {ori_name: name, name: ori_name}"""
        self.node_name_map = {}
        for node in self.nodes():
            if isinstance(node, (BoundInput, BoundParams)):
                for p in list(node.named_parameters()):
                    if node.ori_name not in self.node_name_map:
                        name = f'{node.name}.{p[0]}'
                        self.node_name_map[node.ori_name] = name
                        self.node_name_map[name] = node.ori_name
                for p in list(node.named_buffers()):
                    if node.ori_name not in self.node_name_map:
                        name = f'{node.name}.{p[0]}'
                        self.node_name_map[node.ori_name] = name
                        self.node_name_map[name] = node.ori_name

    # convert a Pytorch model to a model with bounds
    def _convert(self, model, global_input):
        if self.verbose:
            logger.info('Converting the model...')

        if not isinstance(global_input, tuple):
            global_input = (global_input,)
        self.num_global_inputs = len(global_input)

        nodesOP, nodesIn, nodesOut, template = self._convert_nodes(
            model, global_input)
        global_input = self._to(global_input, self.device)

        while True:
            self._build_graph(nodesOP, nodesIn, nodesOut, template)
            self.forward(*global_input)  # running means/vars changed
            nodesOP, nodesIn, finished = self._split_complex(nodesOP, nodesIn)
            if finished:
                break

        self._get_node_name_map()

        # Load self.ori_state_dict again to avoid the running means/vars changed
        # during forward().
        self.load_state_dict(self.ori_state_dict)
        if self.ori_training:
            model.load_state_dict(self.ori_state_dict)
        delattr(self, 'ori_state_dict')

        # The final node used in the last time calling `compute_bounds`
        self.last_final_node = None
        self.used_nodes = []

        if self.verbose:
            logger.info('Model converted to support bounds')

    def check_prior_bounds(self, node):
        if node.prior_checked or not (node.used and node.perturbed):
            return
        for n in node.inputs:
            self.check_prior_bounds(n)
        for i in range(len(node.inputs)):
            if (i in node.requires_input_bounds or not node.inputs[i].perturbed
                    or node.inputs[i].name in self.layers_with_constraint):
                self.compute_intermediate_bounds(
                    node.inputs[i], prior_checked=True)
        node.prior_checked = True

    def compute_intermediate_bounds(self, node, prior_checked=False):
        if getattr(node, 'lower', None) is not None:
            if node.name in self.layers_with_constraint:
                node.clamp_interim_bounds()
            return

        logger.debug(f'Getting the bounds of {node}')

        if not prior_checked:
            self.check_prior_bounds(node)

        if not node.perturbed:
            fv = self.get_forward_value(node)
            node.interval = node.lower, node.upper = fv, fv
            return

        # FIXME check that weight perturbation is not affected
        #      (from_input=True should be set for weights)
        if not node.from_input and hasattr(node, 'forward_value'):
            node.lower = node.upper = self.get_forward_value(node)
            return

        reference_bounds = self.reference_bounds

        if self.use_forward:
            node.lower, node.upper = self.forward_general(node=node, concretize=True)
            return

        #FIXME need clean up

        # assign concretized bound for ReLU layer to save computational cost
        # FIXME: Put ReLU after reshape will cause problem!
        if self.check_IBP_intermediate(node):
            # Intermediate bounds for some operators are directly
            # computed from their input nodes by IBP
            # (such as BoundRelu, BoundNeg)
            logger.debug('IBP propagation for intermediate bounds on %s', node)
        else:
            # For the first linear layer, IBP can give the same tightness
            # as CROWN.
            if not self.check_IBP_first_linear(node):
                sparse_intermediate_bounds_with_ibp = self.bound_opts.get('sparse_intermediate_bounds_with_ibp', True)
                # Sparse intermediate bounds can be enabled
                # if aux_reference_bounds are given.
                # (this is enabled for ReLU only, and not for other
                # activations.)
                sparse_intermediate_bounds = (self.bound_opts.get('sparse_intermediate_bounds', False)
                    and isinstance(self[node.output_name[0]], BoundRelu))

                ref_intermediate_lb, ref_intermediate_ub = None, None
                if sparse_intermediate_bounds:
                    if node.name not in self.aux_reference_bounds:
                        # If aux_reference_bounds are not available,
                        # we can use IBP to compute these bounds.
                        if sparse_intermediate_bounds_with_ibp:
                            with torch.no_grad():
                                # Get IBP bounds for this layer;
                                # we set delete_bounds_after_use=True which does
                                # not save extra intermediate bound tensors.
                                ret_ibp = self.IBP_general(node=node, delete_bounds_after_use=True)
                                ref_intermediate_lb = ret_ibp[0]
                                ref_intermediate_ub = ret_ibp[1]
                        else:
                            sparse_intermediate_bounds = False
                    else:
                        aux_bounds = self.aux_reference_bounds[node.name]
                        ref_intermediate_lb, ref_intermediate_ub = aux_bounds

                sparse_C = self.get_sparse_C(
                    node, sparse_intermediate_bounds,
                    ref_intermediate_lb, ref_intermediate_ub)
                newC, reduced_dim, unstable_idx, unstable_size = sparse_C

                if unstable_idx is None or unstable_size > 0:
                    # Special case for BoundRelu when sparse intermediate bounds are disabled
                    # Currently sparse intermediate bounds are restricted to ReLU models only
                    skip = False
                    if unstable_idx is None:
                        if (len(node.output_name) == 1
                                and isinstance(self[node.output_name[0]], (BoundRelu, BoundSignMerge))
                                and node.name in self.reference_bounds):
                            lower, upper = self.reference_bounds[node.name]
                            fully_stable = torch.logical_or(lower>=0, upper<=0).all()
                            if fully_stable:
                                node.lower, node.upper = lower, upper
                                skip = True
                    if not skip:
                        if self.return_A:
                            node.lower, node.upper, _ = self.backward_general(bound_node=node, C=newC, unstable_idx=unstable_idx)
                        else:
                            # Compute backward bounds only when there are unstable
                            # neurons, or when we don't know which neurons are unstable.
                            node.lower, node.upper = self.backward_general(bound_node=node, C=newC, unstable_idx=unstable_idx)

                if reduced_dim:
                    self.restore_sparse_bounds(
                        node, unstable_idx, unstable_size,
                        ref_intermediate_lb, ref_intermediate_ub)

        # node.lower and node.upper (intermediate bounds) are computed in
        # the above function. If we have bound references, we set them here
        # to always obtain a better set of bounds.
        if node.name in reference_bounds:
            ref_bounds = reference_bounds[node.name]
            # Initially, the reference bound and the computed bound can be
            # exactly the same when intermediate layer beta is 0. This will
            # prevent gradients flow. So we need a small guard here.
            # Set the intermediate layer bounds using reference bounds,
            # always choosing the tighter one.
            node.lower = (torch.max(ref_bounds[0], node.lower).detach()
                          - node.lower.detach() + node.lower)
            node.upper = (node.upper - (node.upper.detach()
                          - torch.min(ref_bounds[1], node.upper).detach()))
            # Otherwise, we only use reference bounds to check which neurons
            # are unstable.

        # prior constraint bounds
        if node.name in self.layers_with_constraint:
            node.clamp_interim_bounds()
        # FIXME (12/28): we should be consistent, and only use
        # node.interval, do not use node.lower or node.upper!
        node.interval = (node.lower, node.upper)

    def merge_A_dict(self, lA_dict, uA_dict):
        merged_A = {}
        for output_node_name in lA_dict:
            merged_A[output_node_name] = {}
            lA_dict_ = lA_dict[output_node_name]
            uA_dict_ = uA_dict[output_node_name]
            for input_node_name in lA_dict_:
                merged_A[output_node_name][input_node_name] = {
                    'lA': lA_dict_[input_node_name]['lA'],
                    'uA': uA_dict_[input_node_name]['uA'],
                    'lbias': lA_dict_[input_node_name]['lbias'],
                    'ubias': uA_dict_[input_node_name]['ubias'],
                }
        return merged_A

    # @torch.compile
    def compute_bounds(
            self, x=None, aux=None, C=None, method='backward', IBP=False,
            forward=False, bound_lower=True, bound_upper=False, reuse_ibp=False,
            reuse_alpha=False, return_A=False, needed_A_dict=None,
            final_node_name=None, average_A=False,
            interm_bounds=None, reference_bounds=None,
            intermediate_constr=None, alpha_idx=None,
            aux_reference_bounds=None, need_A_only=False,
            decision_thresh=None,
            update_mask=None):
        r"""Main function for computing bounds.

        Args:
            x (tuple or None): Input to the model. If it is None, the input
            from the last `forward` or `compute_bounds` call is reused.
            Otherwise: the number of elements in the tuple should be
            equal to the number of input nodes in the model, and each element in
            the tuple corresponds to the value for each input node respectively.
            It should look similar as the `global_input` argument when used for
            creating a `BoundedModule`.

            aux (object, optional): Auxliary information that can be passed to
            `Perturbation` classes for initializing and concretizing bounds,
            e.g., additional information for supporting synonym word subsitution
            perturbaiton.

            C (Tensor): The specification matrix that can map the output of the
            model with an additional linear layer. This is usually used for
            maping the logits output of the model to classification margins.

            method (str): The main method for bound computation. Choices:
                * `IBP`: purely use Interval Bound Propagation (IBP) bounds.
                * `CROWN-IBP`: use IBP to compute intermediate bounds,
                but use CROWN (backward mode LiRPA) to compute the bounds of the
                final node.
                * `CROWN`: purely use CROWN to compute bounds for intermediate
                nodes and the final node.
                * `Forward`: purely use forward mode LiRPA.
                * `Forward+Backward`: use forward mode LiRPA for intermediate
                nodes, but further use CROWN for the final node.
                * `CROWN-Optimized` or `alpha-CROWN`: use CROWN, and also
                optimize the linear relaxation parameters for activations.
                * `forward-optimized`: use forward bounds with optimized linear
                relaxation.

            IBP (bool, optional): If `True`, use IBP to compute the bounds of
            intermediate nodes. It can be automatically set according to
            `method`.

            forward (bool, optional): If `True`, use the forward mode bound
            propagation to compute the bounds of intermediate nodes. It can be
            automatically set according to `method`.

            bound_lower (bool, default `True`): If `True`, the lower bounds of
            the output needs to be computed.

            bound_upper (bool, default `True`): If `True`, the upper bounds of
            the output needs to be computed.

            reuse_ibp (bool, optional): If `True` and `method` is None, reuse
            the previously saved IBP bounds.

            final_node_name (str, optional): Set the final node in the
            computational graph for bound computation. By default, the final
            node of the originally built computational graph is used.

            return_A (bool, optional): If `True`, return linear coefficients
            in bound propagation (`A` tensors) with `needed_A_dict` set.

            needed_A_dict (dict, optional): A dictionary specifying linear
            coefficients (`A` tensors) that are needed and should be returned.
            Each key in the dictionary is the name of a starting node in
            backward bound propagation, with a list as the value for the key,
            which specifies the names of the ending nodes in backward bound
            propagation, and the linear coefficients of the starting node w.r.t.
            the specified ending nodes are returned. By default, it is empty.

            reuse_alpha (bool, optional): If `True`, reuse previously saved
            alpha values when they are not being optimized.

            decision_thresh (float, optional): In CROWN-optimized mode, we will
            use this decision_thresh to dynamically optimize those domains that
            <= the threshold.

            interm_bounds: A dictionary of 2-element tuple/list
            containing lower and upper bounds for intermediate layers.
            The dictionary keys should include the names of the layers whose
            bounds should be set without recomputation. The layer names can be
            viewed by setting environment variable AUTOLIRPA_DEBUG_GRAPH=1.
            The values of each dictionary elements are (lower_bounds,
            upper_bounds) where "lower_bounds" and "upper_bounds" are two
            tensors with the same shape as the output shape of this layer. If
            you only need to set intermediate layer bounds for certain layers,
            then just include these layers' names in the dictionary.

            reference_bounds: Format is similar to "interm_bounds".
            However, these bounds are only used as a reference, and the bounds
            for intermediate layers will still be computed (e.g., using CROWN,
            IBP or other specified methods). The computed bounds will be
            compared to "reference_bounds" and the tighter one between the two
            will be used.

            aux_reference_bounds: Format is similar to intermediate layer
            bounds. However, these bounds are only used for determine which
            neurons are stable and which neurons are unstable for ReLU networks.
            Unstable neurons' intermediate layer bounds will be recomputed.

        Returns:
            bound (tuple): When `return_A` is `False`, return a tuple of
            the computed lower bound and upper bound. When `return_A`
            is `True`, return a tuple of lower bound, upper bound, and
            `A` dictionary.
        """
        # This method only prepares everything by setting all required parameters.
        # The main logic is located in `_compute_bounds_main`. It may be called
        # repeatedly for CROWN optimizations.
        logger.debug(f'Compute bounds with {method}')

        if needed_A_dict is None: needed_A_dict = {}
        if not bound_lower and not bound_upper:
            raise ValueError(
                'At least one of bound_lower and bound_upper must be True')

        # Several shortcuts.
        compute_optimized = False
        method = method.lower() if method is not None else method
        if method == 'ibp':
            # Pure IBP bounds.
            method, IBP = None, True
        elif method in ['ibp+backward', 'ibp+crown', 'crown-ibp']:
            method, IBP = 'backward', True
        elif method == 'crown':
            method = 'backward'
        elif method == 'forward':
            forward = True
        elif method == 'forward+backward' or method == 'forward+crown':
            method, forward = 'backward', True
        elif method in ['crown-optimized', 'alpha-crown', 'forward-optimized']:
            # Lower and upper bounds need two separate rounds of optimization.
            if method == 'forward-optimized':
                method = 'forward'
            else:
                method = 'backward'
            compute_optimized = True

        if reference_bounds is None:
            reference_bounds = {}
        if aux_reference_bounds is None:
            aux_reference_bounds = {}

        # If y in self.backward_node_pairs[x], then node y is visited when
        # doing backward bound propagation starting from node x.
        self.backward_from = dict([(node, []) for node in self._modules])

        if not bound_lower and not bound_upper:
            raise ValueError(
                'At least one of bound_lower and bound_upper in compute_bounds '
                'should be True')
        A_dict = {} if return_A else None

        if x is not None:
            if isinstance(x, torch.Tensor):
                x = (x,)
            self.set_input(*x, interm_bounds=interm_bounds)

        roots = self.roots()
        batch_size = roots[0].value.shape[0]
        dim_in = 0

        for i in range(len(roots)):
            value = roots[i].forward()
            if getattr(roots[i], 'perturbation', None) is not None:
                ret_init = roots[i].perturbation.init(
                    value, aux=aux, forward=forward)
                roots[i].linear, roots[i].center, roots[i].aux = ret_init
                # This input/parameter has perturbation.
                # Create an interval object.
                roots[i].interval = Interval(
                    roots[i].linear.lower, roots[i].linear.upper,
                    ptb=roots[i].perturbation)
                if forward:
                    roots[i].dim = roots[i].linear.lw.shape[1]
                    dim_in += roots[i].dim
            else:
                # This input/parameter does not has perturbation.
                # Use plain tuple defaulting to Linf perturbation.
                roots[i].interval = (value, value)
                roots[i].forward_value = roots[i].value = value
                roots[i].center = roots[i].lower = roots[i].upper = value

            roots[i].lower, roots[i].upper = roots[i].interval

        if forward:
            self.init_forward(roots, dim_in)

        for n in self.nodes():
            if isinstance(n, BoundRelu):
                for node in n.inputs:
                    if isinstance(node, BoundConv):
                        # whether this Conv is followed by a ReLU
                        node.relu_followed = True

            # Inject update mask inside the activations
            # update_mask: None or bool tensor([batch_size])
            # If set to a tensor, only update the alpha and beta of selected
            # element (with element=1).
            n.alpha_beta_update_mask = update_mask

        final = (self.final_node() if final_node_name is None
                 else self[final_node_name])
        # BFS to find out whether each node is used given the current final node
        self._set_used_nodes(final)

        # FIXME clean
        self.use_forward = forward
        self.batch_size = batch_size
        self.dim_in = dim_in
        self.return_A = return_A
        self.A_dict = A_dict
        self.needed_A_dict = needed_A_dict
        self.intermediate_constr = intermediate_constr
        self.reference_bounds = reference_bounds
        self.aux_reference_bounds = aux_reference_bounds
        self.final_node_name = final.name

        if compute_optimized:
            kwargs = dict(x=x, C=C, method=method, interm_bounds=interm_bounds,
                reference_bounds=reference_bounds, return_A=return_A,
                aux_reference_bounds=aux_reference_bounds,
                needed_A_dict=needed_A_dict,
                final_node_name=final_node_name,
                decision_thresh=decision_thresh)
            if bound_upper:
                ret2 = self._get_optimized_bounds(bound_side='upper', **kwargs)
            if bound_lower:
                ret1 = self._get_optimized_bounds(bound_side='lower', **kwargs)
            if bound_lower and bound_upper:
                if return_A:
                    # Needs to merge the A dictionary.
                    return ret1[0], ret2[1], self.merge_A_dict(ret1[2], ret2[2])
                else:
                    return ret1[0], ret2[1]
            elif bound_lower:
                return ret1  # ret1[1] is None.
            elif bound_upper:
                return ret2  # ret2[0] is None.


        return self._compute_bounds_main(C=C,
                                         method=method,
                                         IBP=IBP,
                                         bound_lower=bound_lower,
                                         bound_upper=bound_upper,
                                         reuse_ibp=reuse_ibp,
                                         reuse_alpha=reuse_alpha,
                                         average_A=average_A,
                                         alpha_idx=alpha_idx,
                                         need_A_only=need_A_only,
                                         update_mask=update_mask)

    def save_intermediate(self, save_path=None):
        r"""A function for saving intermediate bounds.

        Please call this function after `compute_bounds`, or it will output
        IBP bounds by default.

        Args:
            save_path (str, default `None`): If `None`, the intermediate bounds
            will not be saved, or it will be saved at the designated path.

        Returns:
            save_dict (dict): Return a dictionary of lower and upper bounds, with
            the key being the name of the layer.
        """
        save_dict = OrderedDict()
        for node in self.nodes():
            if not hasattr(node, 'interval'):
                ibp_lower, ibp_upper = self.IBP_general(node,
                    delete_bounds_after_use=True)
                dim_output = int(prod(node.output_shape[1:]))
                C = torch.eye(dim_output, device=self.device).expand(
                    self.batch_size, dim_output, dim_output)
                crown_lower, crown_upper = self.backward_general(node, C=C)
                save_dict[node.name] = (
                    torch.max(crown_lower, ibp_lower),
                    torch.min(crown_upper, ibp_upper))
            else:
                save_dict[node.name] = (node.lower, node.upper)

        if save_path is not None:
            torch.save(save_dict, save_path)
        return save_dict

    def _compute_bounds_main(self, C=None, method='backward', IBP=False,
            bound_lower=True, bound_upper=True, reuse_ibp=False,
            reuse_alpha=False, average_A=False, alpha_idx=None,
            need_A_only=False, update_mask=None):
        """The core implementation of compute_bounds.

        Seperated because compute_bounds may call _get_optimized_bounds which
        repeatedly calls this method. Otherwise, the preprocessing done in
        compute_bounds would be executed for each iteration.
        """

        final = (self.final_node() if self.final_node_name is None
                 else self[self.final_node_name])
        logger.debug(f'Final node {final.__class__.__name__}({final.name})')

        if IBP and method is None and reuse_ibp:
            # directly return the previously saved ibp bounds
            return self.ibp_lower, self.ibp_upper

        if IBP:
            self.ibp_lower, self.ibp_upper = self.IBP_general(node=final, C=C)

        if method is None:
            return self.ibp_lower, self.ibp_upper

        # TODO: if compute_bounds is called with a method that causes alphas to be
        # optimized, C will be allocated in each iteration. We could allocate it once
        # in compute_bounds, but e.g. `IBP_general` and code in `_get_optimized_bounds`
        # relies on the fact that it can be None
        if C is None:
            # C is an identity matrix by default
            if final.output_shape is None:
                raise ValueError(
                    f'C is not missing while node {final} has no default shape')
            dim_output = int(prod(final.output_shape[1:]))
            # TODO: use an eyeC object here.
            C = torch.eye(dim_output, device=self.device).expand(
                self.batch_size, dim_output, dim_output)

        # Reuse previously saved alpha values,
        # even if they are not optimized now
        # This must be done here instead of `compute_bounds`, as other code might change
        # it (e.g. `_get_optimized_bounds`)
        if reuse_alpha:
            self.opt_reuse()
        else:
            self.opt_no_reuse()

        for node in self.nodes():
            # All nodes may need to be recomputed
            node.prior_checked = False

        self.check_prior_bounds(final)

        if method == 'backward':
            # This is for the final output bound.
            # No need to pass in intermediate layer beta constraints.
            ret = self.backward_general(
                bound_node=final, C=C,
                bound_lower=bound_lower, bound_upper=bound_upper,
                average_A=average_A, need_A_only=need_A_only,
                unstable_idx=alpha_idx, update_mask=update_mask,
            )
            # FIXME when C is specified, lower and upper should not be saved to
            # final.lower and final.upper, because they are not the bounds for
            # the node.
            final.lower, final.upper = ret[0], ret[1]
            return ret
        elif method == 'forward':
            return self.forward_general(C=C, node=final, concretize=True)
        else:
            raise NotImplementedError

    def _set_used_nodes(self, final):
        if final.name != self.last_final_node:
            self.last_final_node = final.name
            self.used_nodes = []
            for i in self.nodes():
                i.used = False
            final.used = True
            queue = deque([final])
            while len(queue) > 0:
                n = queue.popleft()
                self.used_nodes.append(n)
                for n_pre in n.inputs:
                    if not n_pre.used:
                        n_pre.used = True
                        queue.append(n_pre)
        # Based on "used" and "perturbed" properties, find out which
        # layer requires intermediate layer bounds.
        self.layers_requiring_bounds = self.get_layers_requiring_bounds()

    from .interval_bound import (
        IBP_general, _IBP_loss_fusion, check_IBP_intermediate,
        check_IBP_first_linear)
    from .forward_bound import (
        forward_general, forward_general_dynamic, forward_refinement, init_forward)
    from .backward_bound import (
        backward_general, get_sparse_C, concretize,
        check_optimized_variable_sparsity, restore_sparse_bounds,
        get_alpha_crown_start_nodes, get_unstable_locations, batched_backward,
        _preprocess_C)
    from .optimized_bounds import (
        _get_optimized_bounds, init_alpha, update_best_beta,
        opt_reuse, opt_no_reuse, _to_float64, _to_default_dtype, get_refined_interm_bounds)
    from .beta_crown import (beta_crown_backward_bound, reset_beta, set_beta, get_split_nodes)
    from .jacobian import (augment_gradient_graph, compute_jacobian_bounds,
                           _expand_jacobian)
    from .optimize_graph import _optimize_graph
    from .edit_graph import add_nodes, add_input_node, delete_node, replace_node


    from .solver_module import (
        build_solver_module, _build_solver_input, _build_solver_general, _reset_solver_vars, _build_solver_refined)
